def j2season(_date, year=686.9799625, marsday=8.5875, startdate=144.61074994):
    """
    Ls date to a KRC season to determine which KRC seasonal lookup tables to use

    Parameters
    -----------
    _date : float
   	        The input date to be converted

   	year : float
   	       The mars year to search within

   	marsday : float
   	          The length of a Mars day

   	startdate : float
   	            The zero, start date

    Returns
    -------
    startseason : int
		  The integer index to the start season
    stopseasons : int
		  The integer index to the stop season
    """
    date = _date
    if date < startdate:
        remainder = (startdate - date) / year
        date = date + int(remainder + 1.0) * year
    dateoffset = (date - startdate) % year
    recordoffset = dateoffset / marsday
    startseason = int(recordoffset)
    stopseason = startseason + 1
    return recordoffset, startseason, stopseason




import pytest

from plio.io import extract_metadata as em
from plio.io import gdal

@pytest.fixture
def wkt_moon():
    wktsrs = 'PROJCS["Moon2000_Mercator180",GEOGCS["GCS_Moon_2000",DATUM["D_Moon_2000",SPHEROID["Moon_2000_IAU_IAG",1737400.0,0.0]],PRIMEM["Reference_Meridian",0.0],UNIT["Degree",0.0174532925199433]],PROJECTION["Mercator"],PARAMETER["False_Easting",0.0],PARAMETER["False_Northing",0.0],PARAMETER["Central_Meridian",180.0],PARAMETER["Standard_Parallel_1",0.0],UNIT["Meter",1.0]]'
    return em.extract_projstring(wktsrs)

@pytest.fixture
def srs_mars():
    wktsrs = 'PROJCS["Mercator",GEOGCS["GCS_Moon_2000",DATUM["D_Moon_2000",SPHEROID["Moon_2000_IAU_IAG",1737400.0,0.0]],PRIMEM["Reference_Meridian",0.0],UNIT["Degree",0.0174532925199433]],PROJECTION["Mercator"],PARAMETER["False_Easting",0.0],PARAMETER["False_Northing",0.0],PARAMETER["Central_Meridian",180.0],PARAMETER["Standard_Parallel_1",0.0],UNIT["Meter",1.0]]'
    return em.extract_projstring(wktsrs)

@pytest.mark.skipif(gdal is None, reason="GDAL not installed")
def test_false_easting(srs_mars):
    e = em.get_false_easting(srs_mars)
    assert e == 0.0

@pytest.mark.skipif(gdal is None, reason="GDAL not installed")
def test_false_northing(srs_mars):
    n = em.get_false_northing(srs_mars)
    assert n == 0.0

@pytest.mark.skipif(gdal is None, reason="GDAL not installed")
def test_projection_name(srs_mars):
    name = em.get_projection_name(srs_mars)
    assert name == 'Mercator_2SP'

@pytest.mark.skipif(gdal is None, reason="GDAL not installed")
def test_axes_extract(srs_mars):
    smajor, sminor, invflattening = em.get_spheroid(srs_mars)
    assert smajor == 1737400.0
    assert sminor == 1737400.0
    assert invflattening == 0.0

@pytest.mark.skipif(gdal is None, reason="GDAL not installed")
def test_get_standard_parallels(srs_mars):
    parallels = em.get_standard_parallels(srs_mars)
    assert parallels[0] == 0.0
    assert parallels[1] == 0.0

@pytest.mark.skipif(gdal is None, reason="GDAL not installed")
def test_get_central_meridian(srs_mars):
    clon = em.get_central_meridian(srs_mars)
    assert clon == 180.0

@pytest.mark.skipif(gdal is None, reason="GDAL not installed")
def test_export_to_proj4(srs_mars):
    """
    Check that proj4 is not supporting Moon2000_Mercator
    """
    proj4 = srs_mars.ExportToProj4()
    for element in '+proj=merc +lon_0=180 +lat_ts=0 +x_0=0 +y_0=0 +a=1737400 +b=1737400 +units=m +no_defs'.split():
        assert element in proj4

@pytest.mark.skipif(gdal is None, reason="GDAL not installed")
def test_scale_factor(srs_mars):
    k = em.get_scale_factor(srs_mars)
    assert k == 1.0

@pytest.mark.skipif(gdal is None, reason="GDAL not installed")
def test_latitude_of_origin(srs_mars):
    lo = em.get_latitude_of_origin(srs_mars)
    assert lo == 90.0



import json
import numpy as np
import datetime

class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)
 
def read_json(inputfile):
    """
    Read the input json file into a python dictionary.

    Parameters
    ==========
    inputfile : str
                PATH to the file on disk

    Returns
    =======
    jobs : dict
           returns a dictionary 
    """
    with open(inputfile, 'r') as f:
        try:
            jdict = json.load(f)
            return jdict
        except IOError: # pragma: no cover
            return


def write_json(outdata, outputfile):
    """
    Write a Python dictionary as a plain-text JSON file

    Parameters
    ==========
    outdata : dict
              The data structure to be serialized
    outputfile : str
                 The file to write the data to.
    """
    try:
        with open(outputfile, 'w') as f:
            f.write(json.dumps(outdata, outputfile))
    except: # pragma: no cover
        raise IOError('Unable to write data to {}'.format(outputfile))



import warnings
import numpy as np

def is_dict_like(value):
    return hasattr(value, 'keys') and hasattr(value, '__getitem__')

def expanded_indexer(key, ndim):
    """Given a key for indexing an ndarray, return an equivalent key which is a
    tuple with length equal to the number of dimensions.
    The expansion is done by replacing all `Ellipsis` items with the right
    number of full slices and then padding the key with full slices so that it
    reaches the appropriate dimensionality.
    """
    if not isinstance(key, tuple):
        # numpy treats non-tuple keys equivalent to tuples of length 1
        key = (key,)
    new_key = []
    # handling Ellipsis right is a little tricky, see:
    # http://docs.scipy.org/doc/numpy/reference/arrays.indexing.html#advanced-indexing
    found_ellipsis = False
    for k in key:
        if k is Ellipsis:
            if not found_ellipsis:
                new_key.extend((ndim + 1 - len(key)) * [slice(None)])
                found_ellipsis = True
            else:
                new_key.append(slice(None))
        else:
            new_key.append(k)
    if len(new_key) > ndim:
        raise IndexError('too many indices')
    new_key.extend((ndim - len(new_key)) * [slice(None)])
    return tuple(new_key)

class _LocIndexer(object):
    def __init__(self, data_array):
        self.data_array = data_array

    def __getitem__(self, key):
        # expand the indexer so we can handle Ellipsis
        key = expanded_indexer(key, 3)
        sl = key[0]
        ifnone = lambda a, b: b if a is None else a
        if isinstance(sl, slice):
            sl = list(range(ifnone(sl.start, 0),
                            ifnone(sl.stop, len(self.data_array.wavelengths)),
                            ifnone(sl.step, 1)))

        if isinstance(sl, (int, float)):
            idx = self._get_idx(sl)
        else:
            idx = [self._get_idx(s) for s in sl]
        key = (idx, key[1], key[2])

        if len(self.data_array.data) != 0:
            return self.data_array.data[key]

        return self.data_array._read(key)

    def _get_idx(self, value, tolerance=2):
        vals = np.abs(self.data_array.wavelengths-value)
        minidx = np.argmin(vals)
        if vals[minidx] >= tolerance:
            warnings.warn("Absolute difference between requested value and found values is {}".format(vals[minidx]))
        return minidx

class _iLocIndexer(object):
    def __init__(self, data_array):
        self.data_array = data_array

    def __getitem__(self, key):
        # expand the indexer so we can handle Ellipsis
        key = expanded_indexer(key, 3)
        sl = key[0]
        ifnone = lambda a, b: b if a is None else a
        if isinstance(sl, slice):
            sl = list(range(ifnone(sl.start, 0),
                            ifnone(sl.stop, len(self.data_array.wavelengths)),
                            ifnone(sl.step, 1)))

        key = (key[0], key[1], key[2])

        if len(self.data_array.data) != 0:
            return self.data_array.data[key]

        return self.data_array._read(key)



import h5py as h5
import numpy as np
import pandas as pd


DEFAULT_COMPRESSION = 'gzip'
DEFAULT_COMPRESSION_VALUE = 8  # 0 - 9


class HDFDataset(h5.File):
    """
    Read / Write an HDF5 dataset using h5py.  If HDF5 is compiled with
    parallel support, this class will support parallel I/O of all supported
    types as well as Pandas dataframes.
    """

    def __init__(self, filename, mode='a'):
        super(HDFDataset, self).__init__(filename, mode)

    def __del__(self):
        self.close()

    @staticmethod
    def df_to_sarray(df):
        """
        Convert a pandas DataFrame object to a numpy structured array.
        This is functionally equivalent to but more efficient than
        np.array(df.to_array())

        From: http://stackoverflow.com/questions/30773073/save-pandas-dataframe-using-h5py-for-interoperabilty-with-other-hdf5-readers

        Parameters
        ----------
        df : dataframe
             the data frame to convert

        Returns
        -------
        z : ndarray
            a numpy structured array representation of df
        """
        v = df.values
        cols = df.columns
        types = [(cols[i], df[k].dtype.type) for (i, k) in enumerate(cols)]
        dtype = np.dtype(types)
        z = np.zeros(v.shape[0], dtype)
        for (i, k) in enumerate(z.dtype.names):
            z[k] = v[:, i]
        return z

    @staticmethod
    def sarray_to_df(sarray, index_column='index'):
        """
        Convert from a structured array back to a Pandas Dataframe

        Parameters
        ----------
        sarray : array
                 numpy structured array

        index_column : str
                       The name of the index column.  Default: 'index'

        Returns
        -------
         : dataframe
           A pandas dataframe
        """

        def remove_field_name(a, name):
            names = list(a.dtype.names)
            if name in names:
                names.remove(name)
            b = a[names]
            return b
        if index_column is not None:
            index = sarray[index_column]
            clean_array = remove_field_name(sarray, 'index')
        else:
            clean_array = sarray
            index = None
        columns = clean_array.dtype.names

        return pd.DataFrame(data=sarray, index=index, columns=columns)



import re
import os
import json
from collections import defaultdict
from functools import singledispatch
from linecache import getline
import numpy as np
import pandas as pd

from plio.utils.utils import is_number, convert_string_to_float

def socetset_keywords_to_dict(keywords, ell=None):
    """
    Convert a SocetCet keywords.list file to JSON

    Parameters
    ----------
    keywords : str
               Path to the socetset keywords.list file or a raw string that
               will be split on '\n' and parsed.
    
    ell : str
          Optional path to the ellipsoid keywords.list file or a raw string 
          that will be split on '\n' and parsed

    Returns
    -------
     data : dict 
            A dictionary containing the socet keywords parsed.

    """
    data = {}

    def parse(lines):
        for l in lines:
            l = l.strip()
            if not l:
                continue
            elems = l.split()
            if is_number(elems[0]) is False:
                key = elems[0]
                if key in data.keys():
                    raise ValueError('Duplicate dictionary key: {}'.format(key))
                data[key] = []
                if len(elems) == 1:
                    continue
                if len(elems) == 2:
                    data[key] = convert_string_to_float(elems[1])
                else:
                    data[key] += [convert_string_to_float(e) for e in elems[1:]]
            else:
                data[key] += [convert_string_to_float(e) for e in elems]

    if os.path.exists(keywords):
        with open(keywords, 'r') as f:
            keywords = f.readlines()
    else:
        keywords = keywords.split('\n')
    parse(keywords)
   
    if ell:
        if os.path.exists(ell):
            with open(ell, 'r') as f:
                ell = f.readlines()
        else:
            ell = ell.split('\n')
        parse(ell)

    return data

@singledispatch
def read_ipf(arg): # pragma: no cover
    return str(arg)

@read_ipf.register(str)
def read_ipf_str(input_data):
    """
    Read a socet ipf file into a pandas data frame

    Parameters
    ----------
    input_data : str
                 path to the an input data file

    Returns
    -------
    df : pd.DataFrame
         containing the ipf data with appropriate column names and indices
    """
    # Check that the number of rows is matching the expected number
    with open(input_data, 'r') as f:
        for i, l in enumerate(f):
            if i == 1:
                cnt = int(l)
            elif i == 2:
                col = l
                break

    columns = np.genfromtxt(input_data, skip_header=2, dtype='unicode',
                            max_rows = 1, delimiter = ',')

    d = [line.split() for line in open(input_data, 'r')]
    d = np.hstack(np.array(d[3:]))

    d = d.reshape(-1, 12).astype('unicode')

    df = pd.DataFrame(d, columns=columns)
    file = os.path.split(os.path.splitext(input_data)[0])[1]
    df['ipf_file'] = pd.Series(np.full((len(df['pt_id'])), file), index = df.index)

    assert int(cnt) == len(df), 'Dataframe length {} does not match point length {}.'.format(int(cnt), len(df))

    # List of data types for columns in Socet set IPF file
    col_dtype = ['str','int32','int32','int32','float64','float64','float64','float64','float64','float64','float64','float64']

    # Build dict of column names and their data types
    dtype_dict = dict(zip(columns, col_dtype))

    # Hard conversion of data types to ensure 'pt_id' is treated as string, 'val', 'fid_val', 'no_obs' flags treated as int
    df = df.astype(dtype_dict)

    return df

@read_ipf.register(list)
def read_ipf_list(input_data_list):
    """
    Read a socet ipf file into a pandas data frame

    Parameters
    ----------
    input_data_list : list
                 list of paths to the a set of input data files

    Returns
    -------
    df : pd.DataFrame
         containing the ipf data with appropriate column names and indices
    """
    frames = []

    for input_file in input_data_list:
        frames.append(read_ipf(input_file))

    df = pd.concat(frames)

    return df

def save_ipf(df, output_path):
    """
    Write a socet ipf file from an ipf-defined pandas dataframe

    Parameters
    ----------
    df          : pd.DataFrame
                  Pandas DataFrame

    output_file : str
                  path to the output data file

    Returns
    -------
    int         : success value
                  0 = success, 1 = errors
    """

    for ipf_file, ipf_df in df.groupby('ipf_file'):

        output_file = os.path.join(output_path, ipf_file + '.ipf')

        # Check that file can be opened
        try:
            outIPF = open(output_file, 'w', newline='\r\n')
        except:
            print('Unable to open output ipf file: {0}'.format(output_file))
            return 1

        #grab number of rows in pandas dataframe ipf group
        numpts = len(ipf_df)

        #Output ipf header
        outIPF.write('IMAGE POINT FILE\n')
        outIPF.write('{0}\n'.format(numpts))
        outIPF.write('pt_id,val,fid_val,no_obs,l.,s.,sig_l,sig_s,res_l,res_s,fid_x,fid_y\n')

        for index, row in ipf_df.iterrows():
            #Output coordinates to ipf file
            outIPF.write('{0} {1} {2} {3}\n'.format(row['pt_id'], row['val'], row['fid_val'], row['no_obs']))
            outIPF.write('{:0.6f}  {:0.6f}\n'.format(row['l.'], row['s.']))
            outIPF.write('{:0.6f}  {:0.6f}\n'.format(row['sig_l'], row['sig_s']))
            outIPF.write('{:0.6f}  {:0.6f}\n'.format(row['res_l'], row['res_s']))
            outIPF.write('{:0.6f}  {:0.6f}\n\n'.format(row['fid_x'], row['fid_y']))

        outIPF.close()
    return

def read_gpf(input_data,gxp=False):
    """
    Read a socet gpf file into a pandas data frame

    Parameters
    ----------
    input_data : str
                 path to the input data file

    gxp        : bool
                 Flag to notify function if input_data follows Socet GXP format.
                 Defaults is False (assume Socet Set format).

    Returns
    -------
    df : pd.DataFrame
         containing the gpf data with appropriate column names and indices
    """

    # line containing point count differs between Socet Set and Socet GXP gpf files
    if not gxp:
        l = 2
    else:
        l = 21

    cnt = getline(input_data, l)
    cnt = int(cnt)
    
    # Lists of column names and their data types
    if not gxp:
        columns = ['point_id','stat','known','lat_Y_North','long_X_East','ht','sig0','sig1','sig2','res0','res1','res2']
        col_dtype = ['str','int32','int32','float64','float64','float64','float64','float64','float64','float64','float64','float64']
    else:
        columns = ['point_id','use','point_type','lat_Y_North','long_X_East','ht','sig0','sig1','sig2','res0','res1','res2',
                   'eigenval0','eigenvec0_i','eigenvec0_j', 'eigenvec0_k',
                   'eigenval1','eigenvec1_i','eigenvec1_j', 'eigenvec1_k',
                   'eigenval2','eigenvec2_i','eigenvec2_j', 'eigenvec2_k']
        col_dtype = ['str','int32','int32','float64','float64','float64','float64','float64','float64','float64','float64','float64',
                     'float64','float64','float64','float64',
                     'float64','float64','float64','float64',
                     'float64','float64','float64','float64']

    # Build dict of column names and their data types
    dtype_dict = dict(zip(columns, col_dtype))

    if not gxp:
        # Mixed types requires read as unicode - let pandas soft convert
        d = np.genfromtxt(input_data, skip_header=(l+1), dtype='unicode')
        d = d.reshape(-1, len(columns))
    else:
        # Read GXP-style GPF a block at a time
        lmax=( (l+1) + ((cnt-1)*8) + 2 )
        for x in range(l+1,lmax,8):
            a = np.genfromtxt(input_data, skip_header=(x), max_rows=4, dtype='unicode')
            b = np.genfromtxt(input_data, skip_header=(x+4), max_rows=3, dtype='unicode')
            if x == (l+1):
                d = np.hstack([np.hstack(a),np.hstack(b)])
            else:
                d = np.vstack(( d, np.hstack([np.hstack(a),np.hstack(b)])) )

    df = pd.DataFrame(d, columns=columns)

    # Hard conversion of data types to ensure 'point_id' is treated as string and 'stat' and 'known' (GXP: 'use', 'point_type')
    #  flags treated as int
    df = df.astype(dtype_dict)

    # Validate the read data with the header point count
    assert int(cnt) == len(df), 'Dataframe length {} does not match point length {}.'.format(int(cnt), len(df))

    return df

def save_gpf(df, output_file):
    """
    Write a socet gpf file from a gpf-defined pandas dataframe

    Parameters
    ----------
    df          : pd.DataFrame
                  Pandas DataFrame

    output_file : str
                  path to the output data file

    Returns
    -------
    int         : success value
                  0 = success, 1 = errors
    """

    # Check that file can be opened
    try:
        outGPF = open(output_file, 'w', newline='\r\n')
    except:
        print('Unable to open output gpf file: {0}'.format(output_file))
        return 1

    #grab number of rows in pandas dataframe
    numpts = len(df)

    #Output gpf header
    outGPF.write('GROUND POINT FILE\n')
    outGPF.write('{0}\n'.format(numpts))
    outGPF.write('point_id,stat,known,lat_Y_North,long_X_East,ht,sig(3),res(3)\n')

    for index,row in df.iterrows():
        #Output coordinates to gpf file
        outGPF.write('{0} {1} {2}\n'.format(row['point_id'], row['stat'], row['known']))
        outGPF.write('{0}         {1}         {2}\n'.format(row['lat_Y_North'], row['long_X_East'], row['ht']))
        outGPF.write('{0} {1} {2}\n'.format(row['sig0'], row['sig1'], row['sig2']))
        outGPF.write('{0} {1} {2}\n\n'.format(row['res0'], row['res1'], row['res2']))

    outGPF.close()
    return

def read_atf(atf_file):
    """
    Reads a .atf file and outputs all of the
    .ipf, .gpf, .sup, .prj, and path to locate the
    .apf file (should be the same as all others)

    Parameters
    ----------
    atf_file : str
               Full path to a socet atf file

    Returns
    -------
    files_dict : dict
                 Dict of files and data associated with a socet
                 project
    """
    with open(atf_file) as f:
        # Extensions of files we want
        files_ext = ['.prj', '.sup', '.ipf', '.gpf']
        files_dict = []
        files = defaultdict(list)

        for line in f:
            ext = os.path.splitext(line)[-1].strip()

            # Check is needed for split as all do not have a space
            if ext in files_ext:

                # If it is the .prj file, it strips the directory away and grabs file name
                if ext == '.prj':
                    files[ext].append(line.strip().split(' ')[1].split('\\')[-1])

                # If the ext is in the list of files we care about, it addes to the dict
                files[ext].append(line.strip().split(' ')[-1])

            else:

                # Adds to the dict even if not in files we care about
                files[ext.strip()].append(line)

        # Gets the base filepath
        files['basepath'] = os.path.dirname(os.path.abspath(atf_file))

        # Creates a dict out of file lists for GPF, PRJ, IPF, and ATF
        files_dict = (dict(files_dict))

        # Sets the value of IMAGE_IPF to all IPF images
        files_dict['IMAGE_IPF'] = files['.ipf']

        # Sets the value of IMAGE_SUP to all SUP images
        files_dict['IMAGE_SUP'] = files['.sup']

        # Sets value for GPF file
        files_dict['GP_FILE'] = files['.gpf'][0]

        # Sets value for PRJ file
        files_dict['PROJECT'] = files['.prj'][0]

        # Sets the value of PATH to the path of the ATF file
        files_dict['PATH'] = files['basepath']

        return files_dict



from enum import IntEnum
from time import gmtime, strftime
import warnings

import pandas as pd
import numpy as np
import pvl
import struct

from plio.io import ControlNetFileV0002_pb2 as cnf
from plio.io import ControlNetFileHeaderV0005_pb2 as cnh5
from plio.io import ControlPointFileEntryV0005_pb2 as cnp5
from plio.utils.utils import xstr, find_in_dict

HEADERSTARTBYTE = 65536
DEFAULTUSERNAME = 'None'


def write_filelist(lst, path="fromlist.lis"):
    """
    Writes a filelist to a file so it can be used in ISIS3.
    Parameters
    ----------
    lst : list
          A list containing full paths to the images used, as strings.
    path : str
           The name of the file to write out. Default: fromlist.lis
    """
    handle = open(path, 'w')
    for filename in lst:
        handle.write(filename)
        handle.write('\n')
    return


class MeasureMessageType(IntEnum):
    """
    An enum to mirror the ISIS3 MeasureLogData enum.
    """
    GoodnessOfFit = 2
    MinimumPixelZScore = 3
    MaximumPixelZScore = 4
    PixelShift = 5
    WholePixelCorrelation = 6
    SubPixelCorrelation = 7

class MeasureLog():

    def __init__(self, messagetype, value):
        """
        A protobuf compliant measure log object.

        Parameters
        ----------
        messagetype : int or str
                      Either the integer or string representation from the MeasureMessageType enum

        value : int or float
                The value to be stored in the message log
        """
        if isinstance(messagetype, int):
            # by value
            self.messagetype = MeasureMessageType(messagetype)
        else:
            # by name
            self.messagetype = MeasureMessageType[messagetype]

        if not isinstance(value, (float, int)):
            raise TypeError(f'{value} is not a numeric type')
        self.value = value

    def __repr__(self):
        return f'{self.messagetype.name}: {self.value}'

    def to_protobuf(self, version=2):
        """
        Return protobuf compliant measure log object representation
        of this class.

        Returns
        -------
        log_message : obj
                      MeasureLogData object suitable to append to a MeasureLog
                      repeated field.
        """
        # I do not see a better way to get to the inner MeasureLogData obj than this
        # imports were not working because it looks like these need to instantiate off
        # an object
        if version == 2:
            log_message = cnf.ControlPointFileEntryV0002().Measure().MeasureLogData()
        elif version == 5:
            log_message = cnp5.ControlPointFileEntryV0005().Measure().MeasureLogData()
        log_message.doubleDataValue = self.value
        log_message.doubleDataType = self.messagetype
        return log_message

    @classmethod
    def from_protobuf(cls, protobuf):
        return cls(protobuf.doubleDataType, protobuf.doubleDataValue)


class IsisControlNetwork(pd.DataFrame):

    # normal properties
    _metadata = ['header']

    @property
    def _constructor(self):
        return IsisControlNetwork

def from_isis(path, remove_empty=True):

    # Now get ready to work with the binary
    with IsisStore(path, mode='rb') as store:
        df = store.read()

    return df

def to_isis(obj, path, mode='wb', version=2,
            headerstartbyte=HEADERSTARTBYTE,
            networkid='None', targetname='None',
            description='None', username=DEFAULTUSERNAME,
            creation_date=None, modified_date=None,
            pointid_prefix=None, pointid_suffix=None):

    if targetname == 'None':
        warnings.warn("Users should provide a targetname to this function such as 'Moon' or 'Mars' in order to generate a valid ISIS control network.")

    with IsisStore(path, mode) as store:
        if not creation_date:
            creation_date = strftime("%Y-%m-%d %H:%M:%S", gmtime())
        if not modified_date:
            modified_date = strftime("%Y-%m-%d %H:%M:%S", gmtime())
        point_messages, point_sizes = store.create_points(obj, pointid_prefix, pointid_suffix)
        points_bytes = sum(point_sizes)
        buffer_header, buffer_header_size = store.create_buffer_header(networkid,
                                                                       targetname,
                                                                       description,
                                                                       username,
                                                                       point_sizes,
                                                                       creation_date,
                                                                       modified_date)
        # Write the buffer header
        store.write(buffer_header, HEADERSTARTBYTE)
        # Then write the points, so we know where to start writing, + 1 to avoid overwrite
        point_start_offset = HEADERSTARTBYTE + buffer_header_size
        for i, point in enumerate(point_messages):
            store.write(point, point_start_offset)
            point_start_offset += point_sizes[i]
        header = store.create_pvl_header(version, headerstartbyte, networkid,
                                         targetname, description, username,
                                         buffer_header_size, points_bytes,
                                         creation_date, modified_date)

        store.write(header.encode('utf-8'))

class IsisStore(object):
    """
    Class to manage IO of an ISIS control network (version 2).
    Attributes
    ----------
    pointid : int
              The current index to be assigned to newly added points
    """

    point_field_map = {
        'type' : 'pointType',
        'chooserName' : 'pointChoosername',
        'datetime' : 'pointDatetime',
        'editLock' : 'pointEditLock',
        'ignore' : 'pointIgnore',
        'jigsawRejected' : 'pointJigsawRejected',
        'log' : 'pointLog'
    }
    measure_field_map = {
        'type' : 'measureType',
        'choosername' : 'measureChoosername',
        'datetime' : 'measureDatetime',
        'editLock' : 'measureEditLock',
        'ignore' : 'measureIgnore',
        'jigsawRejected' : 'measureJigsawRejected',
        'log' : 'measureLog'
    }

    def __init__(self, path, mode=None, **kwargs):
        self.nmeasures = 0
        self.npoints = 0

        # Conversion from buffer types to Python types
        bt = {1: float,
              5: int,
              8: bool,
              9: str,
              11: list,
              14: int}
        self.header_attrs = [(i.name, bt[i.type]) for i in cnf._CONTROLNETFILEHEADERV0002.fields]
        self.point_attrs = [(i.name, bt[i.type]) for i in cnf._CONTROLPOINTFILEENTRYV0002.fields]
        self.measure_attrs = [(i.name, bt[i.type]) for i in cnf._CONTROLPOINTFILEENTRYV0002_MEASURE.fields]

        self._path = path
        if not mode:
            mode = 'a' # pragma: no cover
        self._mode = mode
        self._handle = None

        self._open()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.close()

    def close(self):
        if self._handle is not None:
            self._handle.close()
        self._handle = None

    def _open(self):
        self._handle = open(self._path, self._mode)

    def read(self):
        """
        Given an ISIS store, read the underlying ISIS3 compatible control network and
        return an IsisControlNetwork dataframe.
        """
        pvl_header = pvl.load(self._path, grammar=pvl.grammar.ISISGrammar())
        header_start_byte = find_in_dict(pvl_header, 'HeaderStartByte')
        header_bytes = find_in_dict(pvl_header, 'HeaderBytes')
        point_start_byte = find_in_dict(pvl_header, 'PointsStartByte')
        version = find_in_dict(pvl_header, 'Version')

        if version == 2:
            self.point_attrs = [i for i in cnf._CONTROLPOINTFILEENTRYV0002.fields_by_name if i != 'measures']
            self.measure_attrs = [i for i in cnf._CONTROLPOINTFILEENTRYV0002_MEASURE.fields_by_name]

            cp = cnf.ControlPointFileEntryV0002()
            self._handle.seek(header_start_byte)
            pbuf_header = cnf.ControlNetFileHeaderV0002()
            pbuf_header.ParseFromString(self._handle.read(header_bytes))

            self._handle.seek(point_start_byte)
            cp = cnf.ControlPointFileEntryV0002()
            pts = []
            for s in pbuf_header.pointMessageSizes:
                cp.ParseFromString(self._handle.read(s))
                pt = [getattr(cp, i) for i in self.point_attrs if i != 'measures']
                for measure in cp.measures:
                    meas = pt + [getattr(measure, j) for j in self.measure_attrs]
                    pts.append(meas)

        elif version == 5:
            self.point_attrs = [i for i in cnp5._CONTROLPOINTFILEENTRYV0005.fields_by_name if i != 'measures']
            self.measure_attrs = [i for i in cnp5._CONTROLPOINTFILEENTRYV0005_MEASURE.fields_by_name]

            cp = cnp5.ControlPointFileEntryV0005()
            self._handle.seek(header_start_byte)
            pbuf_header = cnh5.ControlNetFileHeaderV0005()
            pbuf_header.ParseFromString(self._handle.read(header_bytes))

            self._handle.seek(point_start_byte)
            cp = cnp5.ControlPointFileEntryV0005()
            pts = []
            byte_count = 0
            while byte_count < find_in_dict(pvl_header, 'PointsBytes'):
                message_size = struct.unpack('I', self._handle.read(4))[0]
                cp.ParseFromString(self._handle.read(message_size))
                pt = [getattr(cp, i) for i in self.point_attrs if i != 'measures']

                for measure in cp.measures:
                    meas = pt + [getattr(measure, j) for j in self.measure_attrs]
                    pts.append(meas)

                byte_count += 4 + message_size

        # Some point and measure fields have the same name, so mangle them as point_ and measure_
        point_cols = [self.point_field_map[attr] if attr in self.point_field_map else attr for attr in self.point_attrs]
        measure_cols = [self.measure_field_map[attr] if attr in self.measure_field_map else attr for attr in self.measure_attrs]
        cols = point_cols + measure_cols
        df = IsisControlNetwork(pts, columns=cols)
        # Convert the (0.5, 0.5) origin pixels back to (0,0) pixels
        df['line'] -= 0.5
        df['sample'] -= 0.5
        if 'aprioriline' in df.columns:
            df['aprioriline'] -= 0.5
            df['apriorisample'] -= 0.5

        # Munge the MeasureLogData into Python objs
        df['measureLog'] = df['measureLog'].apply(lambda x: [MeasureLog.from_protobuf(i) for i in x])

        df.header = pvl_header
        return df


    def write(self, data, offset=0):
        """
        Parameters
        ----------
        data : bytes
               Encoded header to be written to the file
        offset : int
                 The byte offset into the output binary
        """
        self._handle.seek(offset)
        self._handle.write(data)

    def create_points(self, df, pointid_prefix, pointid_suffix):
        """
        Step through a control network (C) and return protocol buffer point objects
        Parameters
        ----------
        df : DataFrame
              with the appropriate attributes: point_id, point_type, serial,
              measure_type, x, y required.
              The entries in the list must support grouping by the point_id attribute.
        Returns
        -------
        point_messages : list
                         of serialized points buffers
        point_sizes : list
                      of integer point sizes
        """
        def _set_pid(pointid):
            return '{}{}{}'.format(xstr(pointid_prefix),
                                   pointid,
                                   xstr(pointid_suffix))

        # TODO: Rewrite using apply syntax for performance
        point_sizes = []
        point_messages = []
        for i, g in df.groupby('id'):

            # Get the point specification from the protobuf
            point_spec = cnf.ControlPointFileEntryV0002()

            # Set the ID and then loop over all of the attributes that the
            # point has and check for corresponding columns in the group and
            # set with the correct type
            #point_spec.id = _set_pid(i)
            point_spec.id = _set_pid(i)
            point_spec.type = g.iloc[0].pointType
            try:
                point_spec.referenceIndex = g.iloc[0].referenceIndex
            except:
                warnings.warn(f'Unable to identify referenceIndex for point {point_spec.id}. Defaulting to index 0.')
                point_spec.referenceIndex = 0
            for attr, attrtype in self.point_attrs:
                # Un-mangle common attribute names between points and measures
                df_attr = self.point_field_map.get(attr, attr)
                if df_attr in g.columns:
                    if df_attr == 'pointLog':
                        # Currently pointLog is not supported.
                        warnings.warn('The pointLog field is currently unsupported. Any pointLog data will not be saved.')
                        continue
                    # As per protobuf docs for assigning to a repeated field.
                    if df_attr == 'aprioriCovar' or df_attr == 'adjustedCovar':
                        arr = g.iloc[0][df_attr]
                        if isinstance(arr, np.ndarray):
                            arr = arr.ravel().tolist()
                        if arr:
                            point_spec.aprioriCovar.extend(arr)
                    # If field is repeated you must extend instead of assign
                    elif cnf._CONTROLPOINTFILEENTRYV0002.fields_by_name[attr].label == 3:
                        getattr(point_spec, attr).extend(g.iloc[0][df_attr])
                    else:
                        setattr(point_spec, attr, attrtype(g.iloc[0][df_attr]))

            # A single extend call is cheaper than many add calls to pack points
            measure_iterable = []
            for node_id, m in g.iterrows():
                measure_spec = point_spec.Measure()
                # For all of the attributes, set if they are an dict accessible attr of the obj.
                for attr, attrtype in self.measure_attrs:
                    # Un-mangle common attribute names between points and measures
                    df_attr = self.measure_field_map.get(attr, attr)
                    if df_attr in g.columns:
                        if df_attr == 'measureLog':
                            [getattr(measure_spec, attr).extend([i.to_protobuf()]) for i in m[df_attr]]
                        # If field is repeated you must extend instead of assign
                        elif cnf._CONTROLPOINTFILEENTRYV0002_MEASURE.fields_by_name[attr].label == 3:
                            getattr(measure_spec, attr).extend(m[df_attr])
                        else:
                            setattr(measure_spec, attr, attrtype(m[df_attr]))
                # ISIS pixels are centered on (0.5, 0.5). NDArrays are (0,0) based.
                measure_spec.sample = m['sample'] + 0.5
                measure_spec.line = m['line'] + 0.5
                if 'apriorisample' in g.columns:
                    measure_spec.apriorisample = m['apriorisample'] + 0.5
                    measure_spec.aprioriline = m['aprioriline'] + 0.5
                measure_iterable.append(measure_spec)
                self.nmeasures += 1

            self.npoints += 1

            point_spec.measures.extend(measure_iterable)
            point_message = point_spec.SerializeToString()
            point_sizes.append(point_spec.ByteSize())
            point_messages.append(point_message)
        return point_messages, point_sizes

    def create_buffer_header(self, networkid, targetname,
                             description, username, point_sizes,
                             creation_date,
                             modified_date):
        """
        Create the Google Protocol Buffer header using the
        protobuf spec.
        Parameters
        ----------
        networkid : str
                    The user defined identifier of this control network
        targetname : str
                 The name of the target, e.g. Moon
        description : str
                  A description for the network.
        username : str
               The name of the user / application that created the control network
        point_sizes : list
                      of the point sizes for each point message
        Returns
        -------
        header_message : str
                  The serialized message to write
        header_message_size : int
                              The size of the serialized header, in bytes
        """
        raw_header_message = cnf.ControlNetFileHeaderV0002()
        raw_header_message.created = creation_date
        raw_header_message.lastModified = modified_date
        raw_header_message.networkId = networkid
        raw_header_message.description = description
        raw_header_message.targetName = targetname
        raw_header_message.userName = username
        raw_header_message.pointMessageSizes.extend(point_sizes)

        header_message_size = raw_header_message.ByteSize()
        header_message = raw_header_message.SerializeToString()

        return header_message, header_message_size

    def create_pvl_header(self, version, headerstartbyte,
                      networkid, targetname, description, username,
                          buffer_header_size, points_bytes,
                          creation_date, modified_date):
        """
        Create the PVL header object
        Parameters
        ----------
        version : int
              The current ISIS version to write, defaults to 2
        headerstartbyte : int
                          The seek offset that the protocol buffer header starts at
        networkid : str
                    The name of the network
        targetname : str
                     The name of the target, e.g. Moon
        description : str
                      A description for the network.
        username : str
                   The name of the user / application that created the control network
        buffer_header_size : int
                             Total size of the header in bytes
        points_bytes : int
                       The total number of bytes all points require
        Returns
        -------
         : object
           An ISIS compliant PVL header object
        """

        encoder = pvl.encoder.ISISEncoder(end_delimiter=False)

        header_bytes = buffer_header_size
        points_start_byte = HEADERSTARTBYTE + buffer_header_size

        header = pvl.PVLModule([
            ('ProtoBuffer',
                ({'Core':{'HeaderStartByte': headerstartbyte,
                        'HeaderBytes': header_bytes,
                        'PointsStartByte': points_start_byte,
                        'PointsBytes': points_bytes},

                  'ControlNetworkInfo': pvl.PVLGroup([
                        ('NetworkId', networkid),
                        ('TargetName', targetname),
                        ('UserName', username),
                        ('Created', creation_date),
                        ('LastModified', modified_date),
                        ('Description', description),
                        ('NumberOfPoints', self.npoints),
                        ('NumberOfMeasures', self.nmeasures),
                        ('Version', version)
                        ])
                  }),

                 )
        ])

        return pvl.dumps(header, encoder=encoder)



import logging
import re
import warnings

import numpy as np
import pandas as pd

idl2np_dtype = {1: np.byte, 2: np.int16, 3: np.int32, 4: np.float32,
                5: np.float64}
idl2struct = {4: 'f', 5:'d'}
archtype2struct={'sparc': None, 'bigend': '>', 'litend': '<',
          'alpha': None, 'ppc': None, 'x86': None, 'x86_64': None}


class ReadBin52(object):
    """
    Class to read a bin 52 and organize the output


    Attributes
    ----------
    bin52data : ndarray
                The raw, n-dimensional KRC data

    casesize : int
               The number of bytes in each KRC case

    date : bytes
           The date the read file was created

    ddd : ndarray
          Raw temperature data

    ddd_pd : dataframe
             A pandas dataframe of surface temperature data

    filename : str
               The name of the KRC file that was input

    header : list
             The KRC header unpack to a list of ints

    headerlength : int
                   The length, in bytes, of the KRC header

    ncases : int
             The number of different cases the model was run for

    ndim : int
           TBD

    ndx : int
          The number of indices for a single KRC run

    nhours : int
             The number of hour bins per 24 Mars hours

    nlats : int
            The number of valid, non-zero latitude bins

    nseasons : int
               The number of seasons the model was run for

    nvariables : int
                 The number of variables contained within the KRC
                 lookup table

    structdtype : str
                  Describing endianess and data type

    temperature_data : dataframe
                       A multi-indexed dataframe of surface temperatures

    transferedlayers : int
                       The number of KRC transfered layers

    version : list
              The krc verison used to create the file in the form
              [major, minor, release]

    words_per_krc : int
                    The number of bytes per krc entry
    """

    def __init__(self, filename, headerlen=512):
        """
        Parameters
        ----------
        filename : str
                   The file to read

        headerlength : int
                       The length, in bytes, of the text header.
                       Default: 512
        """
        # Get or setup the logging object
        self.logger = logging.getLogger(__name__)

        self.filename = filename
        self.readbin5(headerlen)
        print(self.ncases)
        assert(self.ncases == self.bin52data.shape[0])

    def definekrc(self, what='KRC', endianness='<'):
        """
        Defines a custom binary data structure for the KRC files.
        """
        if what == 'KRC':
            numfd = 96  # Size of floats of T-dependent materials
            numid = 40  # size of "  " integers
            numld = 20  # size of "  " logicals
            maxn1 = 30  # dimension of layers
            maxn2 = 384 * 4  # dimensions of times of day
            maxn3 = 16  # dimensions of iterations days
            maxn4 = self.nlats * 2 - 1  # dimensions of latitudes
            maxn5 = 161 # dimensions of seasons
            maxn6 = 6  # dimensions of saved years
            maxnh = self.nhours  # dimensions of saved times of day
            maxbot = 6  # dimensions of time divisions
            numtit = 20  # number of 4-byte words in TITLE
            numday = 5  # number of 4-byte words in DAY
            e = endianness
            self.logger.debug(self.structdtype)

            #Define the structure of the KRC file
            if self.structdtype == '<f':
                self._krcstructure= np.dtype([('fd','{}{}f'.format(e, numfd)),
                                        ('id','{}{}i'.format(e, numid)),
                                        ('ld','{}{}i'.format(e, numld)),
                                        ('title','{}{}a'.format(e, 4 * numtit)),
                                        ('daytime','{}{}a'.format(e, 4 * numday)),
                                        ('alat','{}{}f4'.format(e, maxn4)),
                                        ('elev','{}{}f4'.format(e,maxn4))])
            elif self.structdtype == '<d':
                self._krcstructure = np.dtype([('fd','{}{}d'.format(e, numfd)),
                                               ('alat','{}{}d'.format(e, maxn4)),
                                               ('elev','{}{}d'.format(e,maxn4) ),
                                               ('id','{}{}i'.format(e, numid)),
                                               ('ld','{}{}i'.format(e, numld)),
                                               ('title','{}{}a'.format(e, 4 * numtit) ),
                                               ('daytime','{}{}a'.format(e, 4 * numday))])

    def readbin5(self, headerlen):
        """
        Reads the type 52 file containing KRC output.

        Tested with KRC version 2.2.2.  Note that the output format
        can change

        Parameters
        ----------
        filename    (str) Full PATH to the file
        """
        def _parse_header():
            header = re.findall(b'\d+', fullheader.split(b'<<')[0])
            header = list(map(int, header))
            self.ndim = header[0]
            self.nhours = header[1]
            self.nvariables = header[2]
            self.nlats = header[3]
            self.nseasons = header[4] - 1
            self.headerlength = header[8]
            self.ncases = header[0 + self.ndim]
            self.header = header
            print(self.header)
            # Compute how large each case is
            self.casesize = self.nhours
            if self.ndim > 1:
                for k in range(1, self.ndim - 1):
                    self.casesize *= header[k + 1]

        def _parse_front():
            # Read the front matter
            front = np.fromfile(bin5, dtype=self.structdtype, count=4).astype(np.int)
            self.words_per_krc = front[0]
            self.ndx = front[2]

        def _read_data():
            bin5.seek(self.headerlength)
            self.bin52data = np.fromfile(bin5,
                                         dtype=self.structdtype)
            self.logger.debug(len(self.bin52data))

            indices = arraysize[1: -1]
            self.bin52data = self.bin52data.reshape(indices[: : -1])

        def _read_metadata():
            if self.structdtype == '<f':
                j = self.headerlength + 16  # Skip header plus 1 16bit entry
            elif self.structdtype == '<d':
                j = self.headerlength + 32 # Skip header plus 1 32bit entry

            bin5.seek(j)
            self.definekrc('KRC')
            structarr = np.fromfile(bin5, dtype=self._krcstructure, count=1)

            if self.structdtype == '<f':
                self.structarr = {'fd': structarr[0][0],
                    'id': structarr[0][1],
                    'ld': structarr[0][2],
                    'title': structarr[0][3],
                    'date': structarr[0][4],
                    'alat': structarr[0][5],
                    'elevation': structarr[0][6]
                    }
            elif self.structdtype == '<d':
                self.structarr = {'fd': structarr[0][0],
                        'alat': structarr[0][1],
                        'elevation': structarr[0][2],
                        'id': structarr[0][3],
                        'ld': structarr[0][4],
                        'title': structarr[0][5],
                        'date':structarr[0][6]}

        def _get_version():
            ver = re.findall(b'\d+', head)
            ver = list(map(int, ver))
            self.version = ver[: 3]

        with open(self.filename, 'rb') as bin5:

            #To handle endianness and architectures
            archbytes = 8
            c_end = 5

            fullheader = bin5.read(headerlen)
            _parse_header()
            print(self.header)
            arraysize = self.header[0: self.ndim + 2]
            arraydtypecode = arraysize[arraysize[0] + 1]
            try:
                arraydtype = idl2np_dtype[arraydtypecode]
                self.logger.debug("Dtype: ", arraydtype)
            except KeyError:
                self.logger.error("Unable to determine input datatype.")

            assert(len(self.header) == self.ndim + 4)

            if self.headerlength > 512:
                warnings.Warn('Expected header to be 512 bytes, is {} bytes'.format(self.headerlength))
                return

            #Get the endianness of the input file and the data type (32 or 64 bit)
            archstart = self.headerlength - (archbytes + c_end)
            archend = self.headerlength - c_end
            encodingarch = fullheader[archstart: archend].rstrip()
            encodingarch = encodingarch.decode()

            self._endianness = archtype2struct[encodingarch]
            self.structdtype = self._endianness + idl2struct[arraydtypecode]

            #Get the date and head debug
            idx2 = fullheader.find(b'>>')
            idx1 = idx2 - 21
            self.date = fullheader[idx1: idx1 + 20]
            head = fullheader[idx2 + 2:  self.headerlength - (archbytes + c_end) - 3 - idx2]
            head = head.rstrip()

            # Parse the header
            _get_version()
            _parse_front()
            _read_data()
            _read_metadata()


    def readcase(self, case=0):
        """
        Read a single dimension (case) from a bin52 file.

        Parameters
        -----------
        case : int
               The case to be extracted
        """

        def latitems2dataframe():
            """
            Converts Latitude items to a dataframe
            """
            columns = ['# Days to Compute Soln.',
                      'RMS Temp. Change on Last Day',
                      'Predicted Final Atmospheric Temp.',
                      'Predicted frost amount, [kg/m^2]',
                      'Frost albedo (at the last time step)',
                      'Mean upward heat flow into soil surface on last day, [W/m^2]']

            # Grab the correct slice from the data cube and reshape
            latitems = layeritems[: ,: ,: ,: ,0: 3].reshape(self.ncases,
                                                     self.nseasons,
                                                     self.nlats, len(columns))

            # Multi-index generation
            idxcount = self.nseasons * self.nlats * self.ncases
            idxpercase = self.nseasons * self.nlats
            caseidx = np.empty(idxcount)
            for c in range(self.ncases):
                start = c * idxpercase
                caseidx[start:start+idxpercase] = np.repeat(c, idxpercase)
            nseasvect = np.arange(self.nseasons)
            seasonidx = np.repeat(np.arange(self.nseasons), self.nlats)
            latidx = np.tile(self.latitudes.values.ravel(), self.nseasons)

            # Pack the dataframe
            self.latitude = pd.DataFrame(latitems.reshape(self.nseasons * self.nlats, -1),
                                        index=[caseidx, seasonidx, latidx],
                                        columns=columns)

            self.latitude.index.names = ['Case','Season' ,'Latitude']

        def layer2dataframe():
            """
            Converts layeritems into
            """
            columns = ['Tmin', 'Tmax']

            ddd = layeritems[: ,: ,: ,: ,3: 3 + self.transferedlayers].reshape(self.ncases,
                                                                   self.nseasons,
                                                                   self.nlats,
                                                                   len(columns),
                                                                   self.transferedlayers)

            idxcount = self.nseasons * self.nlats * self.transferedlayers * self.ncases
            caseidx = np.empty(idxcount)
            idxpercase = self.nseasons * self.nlats * self.transferedlayers
            for c in range(self.ncases):
                start = c * idxpercase
                caseidx[start:start + idxpercase] = np.repeat(c, idxpercase)

            seasonidx = np.repeat(np.arange(self.nseasons), idxcount / self.nseasons / self.ncases)
            nlatidx = np.repeat(self.latitudes.values.ravel(), idxcount / self.transferedlayers / self.ncases)
            tranlayeridx = np.tile(np.repeat(np.arange(self.transferedlayers), self.nlats), self.nseasons)
            self.layers = pd.DataFrame(ddd.reshape(idxcount, -1),
                                       columns=columns,
                                       index=[caseidx, seasonidx, nlatidx, tranlayeridx])
            self.layers.index.names = ['Case', 'Season', 'Latitude', 'Layer']

        def latelv2dataframes():
            """
            Convert the latitude and elevation arrays to dataframes
            """
            #All latitudes
            #Hugh made some change to the krcc format, but I cannot find documentation...
            if self.structdtype == '<f':
                alllats = krcc[:,prelatwords:].reshape(2, nlat_include_null, self.ncases)
            elif self.structdtype == '<d':
                alllats = krcc[:,96:170].reshape(2, nlat_include_null, self.ncases)
                #Latitudes and elevations for each case
            latelv = alllats[: ,0: nlat]
            if latelv.shape[-1] == 1:
                latelv = latelv[:,:,0]
            self.latitudes = pd.DataFrame(latelv[0], columns=['Latitude'])
            self.elevations = pd.DataFrame(latelv[1], columns=['Elevation'])

        def season2dataframe():
            columns = ['Current Julian date (offset from J2000.0)',
                 'Seasonal longitude of Sun (in degrees)',
                 'Current surface pressure at 0 elevation (in Pascals)',
                 'Mean visible opacity of dust, solar wavelengths',
                 'Global average columnar mass of frost [kg /m^2]']

            # Build a dataframe of the seasonal information
            seasitems = header[:, 4 + self.words_per_krc: k ].reshape(self.ncases,
                                                                      len(columns),
                                                                      self.nseasons)
            caseidx = np.repeat(np.arange(self.ncases), self.nseasons)
            seasonidx = np.repeat(np.arange(self.nseasons), self.ncases)
            flt_seasitems = seasitems.reshape(len(columns),
                                              self.ncases * self.nseasons)
            self.seasons = pd.DataFrame(flt_seasitems.T,
                                          index=[caseidx,seasonidx],
                                          columns=columns)
            self.seasons.index.names = ['Case', 'Season']

        def hourly2dataframe():
            """
            Converts the hourly 'ttt' vector to a
            labelled Pandas dataframe.
            """
            columns = ['Final Hourly Surface Temp.',
                     'Final Hourly Planetary Temp.',
                     'Final Hourly Atmospheric Temp.',
                     'Hourly net downward solar flux [W/m^2]',
                     'Hourly net downward thermal flux [W/m^2]']
            ttt = self.bin52data[: ,self.ndx: ,: ,0: len(columns),: ].reshape(self.ncases,
                                                                         self.nseasons,
                                                                         self.nlats,
                                                                         len(columns),
                                                                         self.nhours)

            reshapettt = np.swapaxes(ttt.reshape(self.ncases * self.nseasons * self.nlats,
                                                 len(columns),
                                                 self.nhours),1,2)
            shp = reshapettt.shape
            reshapettt = reshapettt.reshape((shp[0] * shp[1], shp[2])).T
            #Indices
            caseidx = np.repeat(np.arange(self.ncases), self.nseasons * self.nlats * self.nhours)
            seasonidx = np.tile(np.repeat(np.arange(self.nseasons), self.nlats * self.nhours), self.ncases)
            latidx = np.tile(np.repeat(self.latitudes.values.ravel(), self.nhours), self.nseasons)
            houridx = np.tile(np.tile(np.tile(np.arange(self.nhours), self.nlats), self.nseasons), self.ncases)

            #DataFrame
            self.temperatures = pd.DataFrame(reshapettt.T,
                                       index=[caseidx, seasonidx, latidx, houridx],
                                       columns=columns)
            self.temperatures.index.names = ['Case', 'Season', 'Latitude', 'Hour']



        self.nlayers = nlayers = self.structarr['id'][0]
        nlat = len(self.structarr['alat'])
        self.transferedlayers = nlayers - 1
        if self.nhours - 3 < self.transferedlayers:
            self.transferedlayers = self.nhours - 3

        wordsperlat = self.nhours * self.nvariables
        wordsperseason = wordsperlat * self.nlats

        # Each case has a header that must be extracted:
        header = self.bin52data[:,0:self.ndx,: ,: ,: ].reshape(self.ncases,
                                                               wordsperseason * self.ndx)
        k = 4 + self.words_per_krc + 5 * self.nseasons
        krcc = header[:, 4: 4 + self.words_per_krc]

        nlat_include_null = len(self.structarr['alat'])
        prelatwords = self.words_per_krc - 2 * nlat_include_null

        wordspercase = wordsperseason * self.nseasons

        # Extract the latitude and elevation metadata
        latelv2dataframes()

        #Extract the hourly temperature data
        hourly2dataframe()

        # Extract the seasons
        season2dataframe()

        # Extract by layer data from the data cube
        layeritems = self.bin52data[: , self.ndx: , : , 5: 7, : ]
        latitems2dataframe()
        layer2dataframe()



class ReadTds(object):

    def __init__(self, filename, headerlength=512):
        """
        Parameters
        ----------
        filename : str
                   The file to read

        headerlength : int
                       The length, in bytes, of the text header.
                       Default: 512
        """
        # Get or setup the logging object
        self.logger = logging.getLogger(__name__)

        self.filename = filename
        self.readbin5(headerlen)
        print(self.ncases)
        assert(self.ncases == self.bin52data.shape[0])



from sqlalchemy import Column, Integer, String, create_engine, orm
from sqlalchemy.ext import declarative

from plio.sqlalchemy_json.alchemy import NestedJsonObject

Base = declarative.declarative_base()


def setup_db_session(db):
    """
    Add a database session object to the root namespace

    Parameters
    ----------
    db : str
         Database name

    Returns
    -------
     : object
       A SQLAlchemy session object
    """
    engine = create_engine('sqlite:///{}'.format(db))
    Base.metadata.bind = engine
    Base.metadata.create_all()
    return orm.sessionmaker(bind=engine)()


class Translations(Base):  # pragma: no cover
    """
    Table mapping for the ISIS Translation file table
    """
    __tablename__ = 'isis_translations'
    id = Column(Integer, primary_key=True)
    mission = Column(String)
    instrument = Column(String)
    translation = Column(NestedJsonObject)

    def __init__(self, mission, instrument, translation):
        self.mission = mission
        self.instrument = instrument
        self.translation = translation

    def __repr__(self):
        return 'Mission: {}\nInstrument: {}\nTranslation: {}'.format(self.mission,
                                                                     self.instrument,
                                                                     self.translation)

class StringToMission(Base):  # pragma: no cover
    """
    Table mapping for the ISIS mission name cleaner table
    """
    __tablename__ = 'isis_mission_to_standard'
    id = Column(Integer, primary_key=True)
    key = Column(String)
    value = Column(String)

    def __init__(self, key, value):
        self.key = key
        self.value = value




import unittest

from plio.examples import get_path
from plio.io import io_edr


# class Test_Tes_IO(unittest.TestCase):
#
#     # Need different test data or need to modify the current code
#     def setUp(self):
#         self.examplefile = get_path('cl5_398736801edr_f0030004ccam01014m1.dat')
#     #
#     def test_open(self):
#         ds = io_edr.EDR(self.examplefile)
#
# if __name__ == '__main__':
#     unittest.main()



import os
import sys

import pytest

from plio.io import isis_serial_number
from plio.examples import get_path

@pytest.fixture
def label(request):
    return get_path(request.param)

@pytest.mark.parametrize("label, expected", [('Test_PVL.lbl', 'APOLLO15/METRIC/1971-07-31T14:02:27.179'),
                                             ('ctx.pvl','MRO/CTX/0906095311:038')
                                             ], indirect=['label'])
def test_generate_serial_number(label, expected):
    serial = isis_serial_number.generate_serial_number(label)
    assert serial == expected

    


from io import BytesIO
import json
import os
import warnings
from zipfile import ZipFile

from networkx.readwrite import json_graph
import numpy as np
import pandas as pd


try:
    import autocnet
    autocnet_avail = True
except:
    autocnet_avail = False

class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        """If input object is an ndarray it will be converted into a dict
        holding dtype, shape and the data, base64 encoded.
        """
        if isinstance(obj, np.ndarray):
            return dict(__ndarray__= obj.tolist(),
                        dtype=str(obj.dtype),
                        shape=obj.shape)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)

def save(network, projectname):
    """
    Save an AutoCNet candiate graph to disk in a compressed file.  The
    graph adjacency structure is stored as human readable JSON and all
    potentially large numpy arrays are stored as compressed binary. The
    project archive is a standard .zip file that can have any ending,
    e.g., <projectname>.project, <projectname>.zip, <projectname>.myname.

    TODO: This func. writes a intermediary .npz to disk when saving.  Can
    we write the .npz to memory?

    Parameters
    ----------
    network : object
              The AutoCNet Candidate Graph object

    projectname : str
                  The PATH to the output file.
    """
    # Convert the graph into json format
    js = json_graph.node_link_data(network)

    with ZipFile(projectname, 'w') as pzip:
        js_str = json.dumps(js, cls=NumpyEncoder, sort_keys=True, indent=4)
        pzip.writestr('graph.json', js_str)

        # Write the array node_attributes to hdf
        for n, data in network.nodes_iter(data=True):
            grp = data['node_id']
            np.savez('{}.npz'.format(data['node_id']),
                     descriptors=data.descriptors,
                     _keypoints=data._keypoints,
                     _keypoints_idx=data._keypoints.index,
                     _keypoints_columns=data._keypoints.columns)
            pzip.write('{}.npz'.format(data['node_id']))
            os.remove('{}.npz'.format(data['node_id']))

        # Write the array edge attributes to hdf
        for s, d, data in network.edges_iter(data=True):
            if s > d:
                s, d = d, s
            grp = str((s,d))
            np.savez('{}_{}.npz'.format(s, d),
                     matches=data.matches,
                     matches_idx=data.matches.index,
                     matches_columns=data.matches.columns,
                     _masks=data._masks,
                     _masks_idx=data._masks.index,
                     _masks_columns=data._masks.columns)
            pzip.write('{}_{}.npz'.format(s, d))
            os.remove('{}_{}.npz'.format(s, d))

def json_numpy_obj_hook(dct):
    """Decodes a previously encoded numpy ndarray with proper shape and dtype.

    :param dct: (dict) json encoded ndarray
    :return: (ndarray) if input was an encoded ndarray
    """
    if isinstance(dct, dict) and '__ndarray__' in dct:
        data = np.asarray(dct['__ndarray__'])
        return np.frombuffer(data, dct['dtype']).reshape(dct['shape'])
    return dct

def load(projectname):
    """
    Read an AutoCNET project into memory.

    Parameters
    ----------
    projectname : str
                  PATH to the project

    Returns
    -------
    cg : object
         AutoCNET CandidateGraph object
    """
    if autocnet_avail is False:
        warning.warn('AutoCNet Library is not available.  Unable to load an AutoCNet CandidateGraph')
        return

    with ZipFile(projectname, 'r') as pzip:
        # Read the graph object
        with pzip.open('graph.json', 'r') as g:
            data = json.loads(g.read().decode(),object_hook=json_numpy_obj_hook)

        cg = autocnet.graph.network.CandidateGraph()
        Edge = autocnet.graph.edge.Edge
        Node = autocnet.graph.node.Node
        # Reload the graph attributes
        cg.graph = data['graph']
        # Handle nodes
        for d in data['nodes']:
            n = Node(image_name=d['image_name'], image_path=d['image_path'], node_id=d['id'])
            n['hash'] = d['hash']
            # Load the byte stream for the nested npz file into memory and then unpack
            nzf = np.load(BytesIO(pzip.read('{}.npz'.format(d['id']))))
            n._keypoints = pd.DataFrame(nzf['_keypoints'], index=nzf['_keypoints_idx'], columns=nzf['_keypoints_columns'])
            n.descriptors = nzf['descriptors']
            cg.add_node(d['node_id'])
            cg.node[d['node_id']] = n
        for e in data['links']:
            cg.add_edge(e['source'], e['target'])
            edge = Edge()
            edge.source = cg.node[e['source']]
            edge.destination = cg.node[e['target']]
            edge['fundamental_matrix'] = e['fundamental_matrix']
            edge['weight'] = e['weight']
            nzf = np.load(BytesIO(pzip.read('{}_{}.npz'.format(e['source'], e['target']))))

            edge._masks = pd.DataFrame(nzf['_masks'], index=nzf['_masks_idx'], columns=nzf['_masks_columns'])
            edge.matches = pd.DataFrame(nzf['matches'], index=nzf['matches_idx'], columns=nzf['matches_columns'])
            # Add a mock edge
            cg.edge[e['source']][e['target']] = edge

    return cg



import numpy as np
from .io_gdal import GeoDataset

def open(input_data):
    ds = GeoDataset(input_data)
    return ds

def getspectra(x, y, ds):
    nbands = ds.RasterCount
    reflectance = np.empty(nbands)
    for b in range(1, nbands + 1):
        reflectance[b - 1] = ds.GetRasterBand(b).ReadAsArray(y, x, 1, 1)

    mergedref = np.empty(nbands - 1)
    mergedref[:4] = reflectance[:4]
    mergedref[4] = (reflectance[4] + reflectance[5]) / 2
    mergedref[5:] = reflectance[6:]
    return mergedref



import warnings

import pvl
from pvl.collections import PVLModule
from itertools import chain

import plio
from plio.data import get_data
from plio.io.io_db import Translations, StringToMission, setup_db_session
from plio.utils.utils import find_in_dict, find_nested_in_dict


def get_isis_translation(label):
    """
    Compute the ISIS serial number for a given image using
    the input cube or the label extracted from the cube.

    Parameters
    ----------
    label : dict or str
            A PVL dict object or file name to extract
            the PVL object from

    Returns
    -------
    translation : dict
                  A PVLModule object containing the extracted
                  translation file
    """
    # Instantiate a DB session if not already instantiated
    if not hasattr(plio, 'data_session'):
        plio.data_session = setup_db_session(get_data('data.db'))

    # Grab the label is not already read
    if not isinstance(label, PVLModule):
        label = pvl.load(label)

    # Grab the spacecraft name and run it through the ISIS lookup
    spacecraft_name = find_in_dict(label, 'SpacecraftName')
    for row in plio.data_session.query(StringToMission).filter(StringToMission.key==spacecraft_name):
        spacecraft_name = row.value.lower()
    # Try and pull an instrument identifier
    try:
        instrumentid = find_in_dict(label, 'InstrumentId').capitalize()
    except:
        instrumentid = None

    translation = None
    # Grab the translation PVL object using the lookup
    for row in plio.data_session.query(Translations).filter(Translations.mission==spacecraft_name,
                                                            Translations.instrument==instrumentid):
        # Convert the JSON back to a PVL object
        translation = PVLModule(row.translation)
    return translation


def generate_serial_number(label):
    """
    Generate an ISIS compatible serial number using the ISIS
    team's translation files

    Parameters
    ----------
    label : dict or str
            A PVLModule object (dict) or string PATH
            to a file containing a PVL header

    Returns
    -------
     : str
       The ISIS compatible serial number
    """
    if not isinstance(label, PVLModule):
        label = pvl.load(label, decoder=SerialNumberDecoder())
    # Get the translation information
    translation = get_isis_translation(label)

    if not translation:
        warnings.warn('Unable to load an appropriate image translation.')
        return

    serial_number = []

    # Sort the keys to ensure proper iteration order
    keys = sorted(translation.keys())

    for k in keys:
        try:
            group = translation[k]
            search_key = group['InputKey']
            search_position = group['InputPosition']
            search_translation = {group['Translation'][1]:group['Translation'][0]}
            sub_group = find_nested_in_dict(label, search_position)
            serial_entry = sub_group[search_key]

            if serial_entry in search_translation.keys():
                serial_entry = search_translation[serial_entry]
            elif '*' in search_translation.keys() and search_translation['*'] != '*':
                serial_entry = search_translation['*']
            serial_number.append(serial_entry)
        except:
            pass

    return '/'.join(serial_number)


class SerialNumberDecoder(pvl.decoder.PVLDecoder):
    """
    A PVL Decoder class to handle cube label parsing for the purpose of creating a valid ISIS
    serial number. Inherits from the PVLDecoder in planetarypy's pvl module.
    """

    def decode_simple_value(self, value: str):
        """Returns a Python object based on *value*, assuming
        that *value* can be decoded as a PVL Simple Value::

         <Simple-Value> ::= (<Numeric> | <String>)

         Modified from https://pvl.readthedocs.io/en/stable/_modules/pvl/decoder.html#PVLDecoder.decode_simple_value
         Modification entails stripping datetime from list of functions.
        """
        for d in (
            self.decode_quoted_string,
            self.decode_non_decimal,
            self.decode_decimal,
        ):
            try:
                return d(value)
            except ValueError:
                pass

        if value.casefold() == self.grammar.none_keyword.casefold():
            return None

        if value.casefold() == self.grammar.true_keyword.casefold():
            return True

        if value.casefold() == self.grammar.false_keyword.casefold():
            return False

        return self.decode_unquoted_string(value)

    def decode_unquoted_string(self, value: str) -> str:
        """Returns a Python ``str`` if *value* can be decoded
        as an unquoted string, based on this decoder's grammar.
        Raises a ValueError otherwise.

        Modified from: https://pvl.readthedocs.io/en/stable/_modules/pvl/decoder.html#PVLDecoder.decode_unquoted_string
        Modification entails removal of decode_datetime call
        """
        for coll in (
            ("a comment", chain.from_iterable(self.grammar.comments)),
            ("some whitespace", self.grammar.whitespace),
            ("a special character", self.grammar.reserved_characters),
        ):
            for item in coll[1]:
                if item in value:
                    raise ValueError(
                        "Expected a Simple Value, but encountered "
                        f'{coll[0]} in "{self}": "{item}".'
                    )

        agg_keywords = self.grammar.aggregation_keywords.items()
        for kw in chain.from_iterable(agg_keywords):
            if kw.casefold() == value.casefold():
                raise ValueError(
                    "Expected a Simple Value, but encountered "
                    f'an aggregation keyword: "{value}".'
                )

        for es in self.grammar.end_statements:
            if es.casefold() == value.casefold():
                raise ValueError(
                    "Expected a Simple Value, but encountered "
                    f'an End-Statement: "{value}".'
                )

        return str(value)



# Third-party modules
try:
  import simplejson as json
except ImportError:
  import json

import sqlalchemy
from sqlalchemy.ext import mutable

# Custom modules
from . import track


class NestedMutable(mutable.MutableDict, track.TrackedDict):
  """SQLAlchemy `mutable` extension dictionary with nested change tracking."""
  def __setitem__(self, key, value):
    """Ensure that items set are converted to change-tracking types."""
    super(NestedMutable, self).__setitem__(key, self.convert(value, self))

  @classmethod
  def coerce(cls, key, value):
    """Convert plain dictionary to NestedMutable."""
    if isinstance(value, cls):
      return value
    if isinstance(value, dict):
      return cls(value)
    return super(cls).coerce(key, value)


class _JsonTypeDecorator(sqlalchemy.TypeDecorator):
  """Enables JSON storage by encoding and decoding on the fly."""
  impl = sqlalchemy.String

  def process_bind_param(self, value, dialect):
    return json.dumps(value)

  def process_result_value(self, value, dialect):
    return json.loads(value)


class JsonObject(_JsonTypeDecorator):
  """JSON object type for SQLAlchemy with change tracking as base level."""


class NestedJsonObject(_JsonTypeDecorator):
  """JSON object type for SQLAlchemy with nested change tracking."""


mutable.MutableDict.associate_with(JsonObject)
NestedMutable.associate_with(NestedJsonObject)



import os
import sys
import unittest

import pandas as pd

sys.path.insert(0, os.path.abspath('..'))

from plio.examples import get_path
from plio.io import io_tes
from plio.io.io_gdal import GeoDataset
from plio.utils._tes2numpy import tes2numpy


class Test_Tes_IO(unittest.TestCase):

    def setUp(self):
        self.examplefile = get_path('pos10001.tab')

    def test_opentab(self):
        ds = io_tes.Tes(self.examplefile)
        self.assertEqual(ds.data.size, 119106)
        self.assertIsInstance(ds.data, pd.DataFrame)
        self.assertEqual(ds.data.columns.tolist(), ['sclk_time', 'et', 'pos', 'sun', 'quat', 'id'])

    def test_tes2numpy(self):
        self.assertEqual(tes2numpy('MSB_UNSIGNED_INTEGER', 4), '>u4')
        self.assertEqual(tes2numpy('MSB_UNSIGNED_INTEGER', 4, 2), [('elem0', '>u4'), ('elem1', '>u4')])

        with self.assertRaises(Exception):
            tes2numpy('IEEE_REAL', 5)

if __name__ == '__main__':
    unittest.main()



import os
import unittest

import numpy as np
import pandas as pd

from plio.io import io_hdf


class TestHDF(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.hdf = io_hdf.HDFDataset('test_io_hdf.hdf', mode='w')
        cls.x = np.array([(0 ,2.,'String'), (1, 3.,"String2")],
                         dtype=[('index', 'i8'),('bar', 'f4'), ('baz', 'O')])
        cls.df = pd.DataFrame(cls.x[['bar', 'baz']], index=cls.x['index'],
                              columns=['bar', 'baz'])

    def test_df_sarray(self):
        converted = self.hdf.df_to_sarray(self.df.reset_index())
        np.testing.assert_array_equal(converted, self.x)

    def test_sarray_df(self):
        converted = self.hdf.sarray_to_df(self.x)
        self.assertTrue((self.df == converted).all().all())

    @classmethod
    def tearDownClass(cls):
        try:
            os.remove('test_io_hdf.hdf')
        except: pass



import os
import json
import logging.config

from ..examples import get_path


def setup_logging(path=get_path('logging.json'),
                  level='INFO',
                  env_key='LOG_CFG'):
    """
    Read a log configuration file, written in JSON

    Parameters
    ----------
    path : string
                   The path to the logging configuration file
    level : object
                    The logger level at which to report
    env_key : str
              A potential environment variable where the user defaults logs
    """
    value = os.getenv(env_key, None)
    if value:
        path = value

    level = getattr(logging, level.upper())
    if os.path.exists(path):
        logtype = os.path.splitext(os.path.basename(path))[1]
        with open(path, 'rt') as f:
            if logtype == '.json':
                config = json.load(f)
            elif logtype == '.yaml':
                import yaml
                config = yaml.safe_load(f.read())
        logging.config.dictConfig(config)
    else:
        logging.basicConfig(level=level)
        logger = logging.getLogger()
        logger.setLevel(level)


import os
import sys
import unittest

import pandas as pd
import pytest

sys.path.insert(0, os.path.abspath('..'))

from plio.examples import get_path
from plio.io import io_spectral_profiler
from plio.io.io_gdal import GeoDataset
from plio.io import gdal


class Test_Spectral_Profiler_IO(unittest.TestCase):

    def setUp(self):
        self.examplefile = get_path('SP_2C_02_02358_S138_E3586.spc')

    def test_openspc(self):
        ds = io_spectral_profiler.Spectral_Profiler(self.examplefile)
        self.assertEqual(ds.nspectra, 38)
        self.assertEqual(ds.spectra[0].columns.tolist(), ['RAW', 'REF1', 'REF2', 'QA', 'RAD'])
    
    @pytest.mark.skipif(gdal is None, reason="GDAL not installed")
    def test_read_browse(self):
        ds = io_spectral_profiler.Spectral_Profiler(self.examplefile)
        ds.open_browse()
        self.assertIsInstance(ds.browse, GeoDataset)
        self.assertEqual(ds.browse.read_array().shape, (512, 456))

class Test_Spectral_Profiler_IO_Detached(unittest.TestCase):
    
    def setUp(self):
        self.examplefile = get_path('SP_2C_03_04184_N187_E0053.spc')
        self.examplelabel = get_path('SP_2C_03_04184_N187_E0053.lbl')
        
    def test_openspc(self):
        ds = io_spectral_profiler.Spectral_Profiler(self.examplefile, self.examplelabel)
        self.assertEqual(ds.nspectra, 38)
        self.assertEqual(ds.spectra[0].columns.tolist(), ['RAW', 'REF1', 'REF2', 'QA', 'RAD'])

if __name__ == '__main__':
    unittest.main()



import numpy as np
import pandas as pd
import pvl

import sys
import functools
import json

from os import path
from plio.io.io_json import read_json
from plio.utils._tes2numpy import tes_dtype_map
from plio.utils._tes2numpy import tes_columns
from plio.utils._tes2numpy import tes_scaling_factors

class Tes(object):
    """
    Attributes
    ----------

    spectra : panel
              A pandas panel containing n individual spectra.

    ancillary_data : dataframe
                     A pandas DataFrame of the parsed ancillary data (PVL label)

    label : object
            The raw PVL label object

    """


    def __init__(self, input_data, var_file = None, data_set=None):
        """
        Read the .spc file, parse the label, and extract the spectra

        Parameters
        ----------

        input_data : string
                     The PATH to the input .tab file

        """
        def expand_column(df, expand_column, columns): # pragma: no cover
            array = np.asarray([np.asarray(list(tup[0])) for tup in df[expand_column].as_matrix()], dtype=np.uint8)
            new_df = pd.concat([df, pd.DataFrame(array, columns=columns)], axis=1)
            del new_df[expand_column]
            return new_df

        def bolquality2arr(arr): # pragma: no cover
            bitarr = np.unpackbits(np.asarray(arr, dtype=np.uint8))
            lis = [(bitarr2int(bitarr[0:3]), bit2bool(bitarr[3:4]))]

            types = [('BOLOMETRIC_INERTIA_RATING', '>u1'), ('BOLOMETER_LAMP_ANOMALY', 'bool_')]
            arr = np.array(lis, dtype=types)
            return arr

        def obsquality2arr(arr): # pragma: no cover
            bitarr = np.unpackbits(np.asarray(arr, dtype=np.uint8))
            lis = [(bitarr2int(bitarr[0:2]), bitarr2int(bitarr[2:5]),
                    bitarr2int(bitarr[5:6]), bitarr2int(bitarr[6:7]),
                    bitarr2int(bitarr[7:8]), bitarr2int(bitarr[8:9]))]

            types = [('HGA_MOTION', '>u1'), ('SOLAR_PANEL_MOTION', '>u1'), ('ALGOR_PATCH', '>u1'),
                     ('IMC_PATCH', '>u1'), ('MOMENTUM_DESATURATION', '>u1'), ('EQUALIZATION_TABLE', '>u1')]
            arr = np.array(lis, dtype=types)
            return arr

        def obsclass2arr(arr): # pragma: no cover
            bitarr = np.unpackbits(np.asarray(arr, dtype=np.uint8))
            lis = [(bitarr2int(bitarr[0:3]), bitarr2int(bitarr[3:7]),
                    bitarr2int(bitarr[7:11]), bitarr2int(bitarr[11:13]),
                    bitarr2int(bitarr[13:14]), bitarr2int(bitarr[14:16]),
                    bitarr2int(bitarr[16:]))]

            types = [('MISSION_PHASE', '>u1'), ('INTENDED_TARGET', '>u1'), ('TES_SEQUENCE', '>u1'),
                     ('NEON_LAMP_STATUS', '>u1'), ('TIMING_ACCURACY', '>u1'), ('SPARE', '>u1'), ('CLASSIFICATION_VALUE', '>u2')]
            arr = np.array(lis, dtype=types)
            return arr

        def radquality2arr(arr): # pragma: no cover
            bitarr = np.unpackbits(np.asarray(arr, dtype=np.uint8))
            lis = [(bitarr2int(bitarr[0:1]), bitarr2int(bitarr[1:2]),
                    bitarr2int(bitarr[2:3]), bitarr2int(bitarr[3:5]),
                    bitarr2int(bitarr[5:7]), bitarr2int(bitarr[5:8]),
                    bitarr2int(bitarr[8:9]))]

            types = [('MAJOR_PHASE_INVERSION', '>u1'), ('ALGOR_RISK', '>u1'), ('CALIBRATION_FAILURE', '>u1'),
                     ('CALIBRATION_QUALITY', '>u1'), ('SPECTROMETER_NOISE', '>u1'), ('SPECTRAL_INERTIA_RATING', '>u1'),
                     ('DETECTOR_MASK_PROBLEM', '>u1')]
            arr = np.array(lis, dtype=types)
            return arr

        def atmquality2arr(arr): # pragma: no cover
            bitarr = np.unpackbits(np.asarray(arr, dtype=np.uint8))
            lis = [(bitarr2int(bitarr[0:2]), bitarr2int(bitarr[2:4]))]

            types = [('TEMPERATURE_PROFILE_RATING', '>u1'), ('ATMOSPHERIC_OPACITY_RATING', '>u1')]
            arr = np.array(lis, dtype=types)
            return arr

        def expand_column(df, expand_column, columns): # pragma: no cover
            array = np.asarray([np.asarray(list(tup[0])) for tup in df[expand_column].as_matrix()], dtype=np.uint8)
            new_df = pd.concat([df, pd.DataFrame(array, columns=columns)], axis=1)
            del new_df[expand_column]
            return new_df

        def bolquality2arr(arr): # pragma: no cover
            bitarr = np.unpackbits(np.asarray(arr, dtype=np.uint8))
            lis = [(bitarr2int(bitarr[0:3]), bit2bool(bitarr[3:4]))]

            types = [('BOLOMETRIC_INERTIA_RATING', '>u1'), ('BOLOMETER_LAMP_ANOMALY', 'bool_')]
            arr = np.array(lis, dtype=types)
            return arr

        def obsquality2arr(arr): # pragma: no cover
            bitarr = np.unpackbits(np.asarray(arr, dtype=np.uint8))
            lis = [(bitarr2int(bitarr[0:2]), bitarr2int(bitarr[2:5]),
                    bitarr2int(bitarr[5:6]), bitarr2int(bitarr[6:7]),
                    bitarr2int(bitarr[7:8]), bitarr2int(bitarr[8:9]))]

            types = [('HGA_MOTION', '>u1'), ('SOLAR_PANEL_MOTION', '>u1'), ('ALGOR_PATCH', '>u1'),
                     ('IMC_PATCH', '>u1'), ('MOMENTUM_DESATURATION', '>u1'), ('EQUALIZATION_TABLE', '>u1')]
            arr = np.array(lis, dtype=types)
            return arr

        def obsclass2arr(arr): # pragma: no cover
            bitarr = np.unpackbits(np.asarray(arr, dtype=np.uint8))
            lis = [(bitarr2int(bitarr[0:3]), bitarr2int(bitarr[3:7]),
                    bitarr2int(bitarr[7:11]), bitarr2int(bitarr[11:13]),
                    bitarr2int(bitarr[13:14]), bitarr2int(bitarr[14:16]),
                    bitarr2int(bitarr[16:]))]

            types = [('MISSION_PHASE', '>u1'), ('INTENDED_TARGET', '>u1'), ('TES_SEQUENCE', '>u1'),
                     ('NEON_LAMP_STATUS', '>u1'), ('TIMING_ACCURACY', '>u1'), ('SPARE', '>u1'), ('CLASSIFICATION_VALUE', '>u2')]
            arr = np.array(lis, dtype=types)
            return arr

        def radquality2arr(arr): # pragma: no cover
            bitarr = np.unpackbits(np.asarray(arr, dtype=np.uint8))
            lis = [(bitarr2int(bitarr[0:1]), bitarr2int(bitarr[1:2]),
                    bitarr2int(bitarr[2:3]), bitarr2int(bitarr[3:5]),
                    bitarr2int(bitarr[5:7]), bitarr2int(bitarr[5:8]),
                    bitarr2int(bitarr[8:9]))]

            types = [('MAJOR_PHASE_INVERSION', '>u1'), ('ALGOR_RISK', '>u1'), ('CALIBRATION_FAILURE', '>u1'),
                     ('CALIBRATION_QUALITY', '>u1'), ('SPECTROMETER_NOISE', '>u1'), ('SPECTRAL_INERTIA_RATING', '>u1'),
                     ('DETECTOR_MASK_PROBLEM', '>u1')]
            arr = np.array(lis, dtype=types)
            return arr

        def atmquality2arr(arr): # pragma: no cover
            bitarr = np.unpackbits(np.asarray(arr, dtype=np.uint8))
            lis = [(bitarr2int(bitarr[0:2]), bitarr2int(bitarr[2:4]))]

            types = [('TEMPERATURE_PROFILE_RATING', '>u1'), ('ATMOSPHERIC_OPACITY_RATING', '>u1')]
            arr = np.array(lis, dtype=types)
            return arr

        def bitarr2int(arr): # pragma: no cover
            arr = "".join(str(i) for i in arr)
            return np.uint8(int(arr,2))

        def bit2bool(bit): # pragma: no cover
            return np.bool_(bit)

        def expand_bitstrings(df, dataset): # pragma: no cover
            if dataset == 'BOL':
                quality_columns = ['ti_bol_rating', 'bol_ref_lamp']
                df['quality'] = df['quality'].apply(bolquality2arr)
                return expand_column(df, 'quality', quality_columns)

            elif dataset == 'OBS':
                quality_columns = ['hga_motion', 'pnl_motion', 'algor_patch', 'imc_patch',
                                   'momentum', 'equal_tab']
                class_columns = ['phase', 'type', 'sequence',
                                 'lamp_status', 'timing', 'spare', 'class_value']

                df['quality'] = df['quality'].apply(obsquality2arr)
                df['class'] = df['class'].apply(obsclass2arr)

                new_df = expand_column(df, 'quality', quality_columns)
                new_df = expand_column(new_df, 'class', class_columns)
                return new_df

            elif dataset == 'RAD':
                quality_columns = ['phase_inversion', 'algor_risk', 'calib_fail', 'calib_quality',
                                   'spect_noise', 'ti_spc_rating', 'det_mask_problem']

                df['quality'] = df['quality'].apply(radquality2arr)

                return expand_column(df, 'quality', quality_columns)

            elif dataset == 'ATM':
                quality_columns = ['atm_pt_rating', 'atm_opacity_rating']

                df['quality'] = df['quality'].apply(atmquality2arr)
                return expand_column(df, 'quality', quality_columns)

            else:
                return df

        if isinstance(input_data, pd.DataFrame):
            self.dataset = None
            if not data_set:
                for key in tes_columns.keys():
                    if len(set(tes_columns[key]).intersection(set(input_data.columns))) > 3 :
                        self.dataset = key
            else:
                self.dataset=data_set

            self.label = None
            self.data = input_data
            return

        self.label = pvl.load(input_data)
        nrecords = self.label['TABLE']['ROWS']
        nbytes_per_rec = self.label['RECORD_BYTES']
        data_start = self.label['LABEL_RECORDS'] * self.label['RECORD_BYTES']
        dataset = self.label['TABLE']['^STRUCTURE'].split('.')[0]
        self.dataset = dataset

        numpy_dtypes = tes_dtype_map
        columns = tes_columns
        scaling_factors = tes_scaling_factors

        with open(input_data, 'rb') as file:
            file.seek(data_start)
            buffer = file.read(nrecords*nbytes_per_rec)
            array = np.frombuffer(buffer, dtype=numpy_dtypes[dataset.upper()]).byteswap().newbyteorder()

        df = pd.DataFrame(data=array, columns=columns[dataset.upper()])

        # Read Radiance array if applicable
        if dataset.upper() == 'RAD': # pragma: no cover
            if not var_file:
                filename, file_extension = path.splitext(input_data)
                var_file = filename + ".var"

            with open(var_file, "rb") as var:
                buffer = var.read()
                def process_rad(index):
                    if index == -1:
                        return None

                    length = np.frombuffer(buffer[index:index+2], dtype='>u2')[0]
                    exp = np.frombuffer(buffer[index+2:index+4], dtype='>i2')[0]
                    scale = 2**(int(exp)-15)
                    radarr = np.frombuffer(buffer[index+4:index+4+length-2], dtype='>i2') * scale
                    if np.frombuffer(buffer[index+4+length-2:index+4+length], dtype='>u2')[0] != length:
                        warnings.warn("Last element did not match the length for file index {} in file {}".format(index, f))
                    return radarr

                df["raw_rad"] = df["raw_rad"].apply(process_rad)
                df["cal_rad"] = df["cal_rad"].apply(process_rad)

        # Apply scaling factors
        for column in scaling_factors[dataset]: # pragma: no cover
            def scale(x):
                 return np.multiply(x, scaling_factors[dataset][column])
            df[column] = df[column].apply(scale)

        df =  expand_bitstrings(df, dataset.upper())

        self.data =  df

    def join(tes_data):
        """
        Given a list of Tes objects, merges them into a single dataframe using
        SPACECRAFT_CLOCK_START_COUNT (sclk_time) as the index.

        Parameters
        ----------

        tes_data : iterable
                   A Python iterable of Tes objects

        Returns
        -------

        : dataframe
          A pandas dataframe containing the merged data

        : outliers
          A list of Tes() objects containing the tables containing no matches
        """
        if not hasattr(tes_data, '__iter__') and not isinstance(tes_data, Tes):
            raise TypeError("Input data must be a Tes datasets or an iterable of Tes datasets, got {}".format(type(tes_data)))
        elif not hasattr(tes_data, '__iter__'):
            tes_data = [tes_data]

        if len(tes_data) == 0:
            warn("Input iterable is empty")

        if not all([isinstance(obj, Tes) for obj in tes_data]):
            # Get the list of types and the indices of elements that caused the error
            types = [type(obj) for obj in tes_data]
            error_idx = [i for i, x in enumerate([isinstance(obj, Tes) for obj in tes_data]) if x == False]

            raise TypeError("Input data must must be a Tes dataset, input array has non Tes objects at indices: {}\
                             for inputs of type: {}".format(error_idx, types))

        single_key_sets = {'ATM', 'POS', 'TLM', 'OBS'}
        compound_key_sets = {'BOL', 'CMP', 'GEO', 'IFG', 'PCT', 'RAD'}
        dfs = dict.fromkeys(single_key_sets | compound_key_sets, DataFrame())

        # Organize the data based on datasets
        for ds in tes_data:
            # Find a way to do this in place?
            dfs[ds.dataset] = dfs[ds.dataset].append(ds.data)

        # remove and dataframes that are empty
        empty_dfs = [key for key in dfs.keys() if dfs[key].empty]
        for key in empty_dfs:
            dfs.pop(key, None)


        single_key_dfs = [dfs[key] for key in dfs.keys() if key in single_key_sets]
        compound_key_dfs = [dfs[key] for key in dfs.keys() if key in compound_key_sets]
        all_dfs = single_key_dfs+compound_key_dfs

        keyspace = functools.reduce(lambda left,right: left|right, [set(df['sclk_time']) for df in all_dfs])

        single_key_merged = functools.reduce(lambda left,right: pd.merge(left, right, on=["sclk_time"]), single_key_dfs)
        compound_key_merged = functools.reduce(lambda left,right: pd.merge(left, right, on=["sclk_time", "detector"]), compound_key_dfs)
        merged = single_key_merged.merge(compound_key_merged, on="sclk_time")

        outlier_idx = keyspace-set(merged["sclk_time"])
        outliers = [Tes(tds.data[tds.data['sclk_time'].isin(outlier_idx)], data_set=tds.dataset) for tds in tes_data]
        return merged, [tds for tds in outliers if not tds.data.empty]



#!/usr/bin/python
"""This module contains the tracked object classes.

TrackedObject forms the basis for both the TrackedDict and the TrackedList.

A function for automatic conversion of dicts and lists to their tracked
counterparts is also included.
"""

# Standard modules
import itertools
import logging


class TrackedObject(object):
  """A base class for delegated change-tracking."""
  _type_mapping = {}

  def __init__(self, *args, **kwds):
    self.logger = logging.getLogger(type(self).__name__)
    self.logger.debug('%s: __init__', self._repr())
    self.parent = None
    super(TrackedObject, self).__init__(*args, **kwds)

  def changed(self, message=None, *args):
    """Marks the object as changed.

    If a `parent` attribute is set, the `changed()` method on the parent will
    be called, propagating the change notification up the chain.

    The message (if provided) will be debug logged.
    """
    if message is not None:
      self.logger.debug('%s: %s', self._repr(), message % args)
    self.logger.debug('%s: changed', self._repr())
    if self.parent is not None:
      self.parent.changed()

  @classmethod
  def register(cls, origin_type):
    """Registers the class decorated with this method as a mutation tracker.

    The provided `origin_type` is mapped to the decorated class such that
    future calls to `convert()` will convert the object of `origin_type` to an
    instance of the decorated class.
    """
    def decorator(tracked_type):
      """Adds the decorated class to the `_type_mapping` dictionary."""
      cls._type_mapping[origin_type] = tracked_type
      return tracked_type
    return decorator

  @classmethod
  def convert(cls, obj, parent):
    """Converts objects to registered tracked types

    This checks the type of the given object against the registered tracked
    types. When a match is found, the given object will be converted to the
    tracked type, its parent set to the provided parent, and returned.

    If its type does not occur in the registered types mapping, the object
    is returned unchanged.
    """
    obj_type = type(obj)
    for origin_type, replacement in cls._type_mapping.items():
      if obj_type is origin_type:
        new = replacement(obj)
        new.parent = parent
        return new
    return obj

  @classmethod
  def convert_iterable(cls, iterable, parent):
    """Returns a generator that performs `convert` on every of its members."""
    return (cls.convert(item, parent) for item in iterable)

  @classmethod
  def convert_iteritems(cls, iteritems, parent):
    """Returns a generator like `convert_iterable` for 2-tuple iterators."""
    return ((key, cls.convert(value, parent)) for key, value in iteritems)

  @classmethod
  def convert_mapping(cls, mapping, parent):
    """Convenience method to track either a dict or a 2-tuple iterator."""
    if isinstance(mapping, dict):
      return cls.convert_iteritems(mapping.items(), parent)
    return cls.convert_iteritems(mapping, parent)

  def _repr(self):
    """Simple object representation."""
    return '<%(namespace)s.%(type)s object at 0x%(address)0xd>' % {
        'namespace': __name__,
        'type': type(self).__name__,
        'address': id(self)}


@TrackedObject.register(dict)
class TrackedDict(TrackedObject, dict):
  """A TrackedObject implementation of the basic dictionary."""
  def __init__(self, source=(), **kwds):
    super(TrackedDict, self).__init__(itertools.chain(
        self.convert_mapping(source, self),
        self.convert_mapping(kwds, self)))

  def __setitem__(self, key, value):
    self.changed('__setitem__: %r=%r', key, value)
    super(TrackedDict, self).__setitem__(key, self.convert(value, self))

  def __delitem__(self, key):
    self.changed('__delitem__: %r', key)
    super(TrackedDict, self).__delitem__(key)

  def clear(self):
    self.changed('clear')
    super(TrackedDict, self).clear()

  def pop(self, *key_and_default):
    self.changed('pop: %r', key_and_default)
    return super(TrackedDict, self).pop(*key_and_default)

  def popitem(self):
    self.changed('popitem')
    return super(TrackedDict, self).popitem()

  def update(self, source=(), **kwds):
    self.changed('update(%r, %r)', source, kwds)
    super(TrackedDict, self).update(itertools.chain(
        self.convert_mapping(source, self),
        self.convert_mapping(kwds, self)))


@TrackedObject.register(list)
class TrackedList(TrackedObject, list):
  """A TrackedObject implementation of the basic list."""
  def __init__(self, iterable=()):
    super(TrackedList, self).__init__(self.convert_iterable(iterable, self))

  def __setitem__(self, key, value):
    self.changed('__setitem__: %r=%r', key, value)
    super(TrackedList, self).__setitem__(key, self.convert(value, self))

  def __delitem__(self, key):
    self.changed('__delitem__: %r', key)
    super(TrackedList, self).__delitem__(key)

  def append(self, item):
    self.changed('append: %r', item)
    super(TrackedList, self).append(self.convert(item, self))

  def extend(self, iterable):
    self.changed('extend: %r', iterable)
    super(TrackedList, self).extend(self.convert_iterable(iterable, self))

  def pop(self, index):
    self.changed('pop: %d', index)
    return super(TrackedList, self).pop(index)

  def sort(self, cmp=None, key=None, reverse=False):
    self.changed('sort')
    super(TrackedList, self).sort(cmp=cmp, key=key, reverse=reverse)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ControlPointFileEntryV0005.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='ControlPointFileEntryV0005.proto',
  package='Isis',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n ControlPointFileEntryV0005.proto\x12\x04Isis\"\x82\x0e\n\x1a\x43ontrolPointFileEntryV0005\x12\n\n\x02id\x18\x01 \x02(\t\x12\x38\n\x04type\x18\x02 \x02(\x0e\x32*.Isis.ControlPointFileEntryV0005.PointType\x12\x13\n\x0b\x63hooserName\x18\x03 \x01(\t\x12\x10\n\x08\x64\x61tetime\x18\x04 \x01(\t\x12\x10\n\x08\x65\x64itLock\x18\x05 \x01(\x08\x12\x0e\n\x06ignore\x18\x06 \x01(\x08\x12\x16\n\x0ejigsawRejected\x18\x07 \x01(\x08\x12\x16\n\x0ereferenceIndex\x18\x08 \x01(\x05\x12N\n\x16\x61prioriSurfPointSource\x18\t \x01(\x0e\x32..Isis.ControlPointFileEntryV0005.AprioriSource\x12\"\n\x1a\x61prioriSurfPointSourceFile\x18\n \x01(\t\x12K\n\x13\x61prioriRadiusSource\x18\x0b \x01(\x0e\x32..Isis.ControlPointFileEntryV0005.AprioriSource\x12\x1f\n\x17\x61prioriRadiusSourceFile\x18\x0c \x01(\t\x12\x1b\n\x13latitudeConstrained\x18\r \x01(\x08\x12\x1c\n\x14longitudeConstrained\x18\x0e \x01(\x08\x12\x19\n\x11radiusConstrained\x18\x0f \x01(\x08\x12\x10\n\x08\x61prioriX\x18\x10 \x01(\x01\x12\x10\n\x08\x61prioriY\x18\x11 \x01(\x01\x12\x10\n\x08\x61prioriZ\x18\x12 \x01(\x01\x12\x18\n\x0c\x61prioriCovar\x18\x13 \x03(\x01\x42\x02\x10\x01\x12\x11\n\tadjustedX\x18\x14 \x01(\x01\x12\x11\n\tadjustedY\x18\x15 \x01(\x01\x12\x11\n\tadjustedZ\x18\x16 \x01(\x01\x12\x19\n\radjustedCovar\x18\x17 \x03(\x01\x42\x02\x10\x01\x12:\n\x03log\x18\x18 \x03(\x0b\x32-.Isis.ControlPointFileEntryV0005.PointLogData\x12:\n\x08measures\x18\x19 \x03(\x0b\x32(.Isis.ControlPointFileEntryV0005.Measure\x1al\n\x0cPointLogData\x12\x16\n\x0e\x64oubleDataType\x18\x01 \x01(\x05\x12\x17\n\x0f\x64oubleDataValue\x18\x02 \x01(\x01\x12\x14\n\x0c\x62oolDataType\x18\x03 \x01(\x05\x12\x15\n\rboolDataValue\x18\x04 \x01(\x08\x1a\x83\x05\n\x07Measure\x12\x14\n\x0cserialnumber\x18\x01 \x02(\t\x12\x42\n\x04type\x18\x02 \x02(\x0e\x32\x34.Isis.ControlPointFileEntryV0005.Measure.MeasureType\x12\x0e\n\x06sample\x18\x03 \x01(\x01\x12\x0c\n\x04line\x18\x04 \x01(\x01\x12\x16\n\x0esampleResidual\x18\x05 \x01(\x01\x12\x14\n\x0clineResidual\x18\x06 \x01(\x01\x12\x13\n\x0b\x63hoosername\x18\x07 \x01(\t\x12\x10\n\x08\x64\x61tetime\x18\x08 \x01(\t\x12\x10\n\x08\x65\x64itLock\x18\t \x01(\x08\x12\x0e\n\x06ignore\x18\n \x01(\x08\x12\x16\n\x0ejigsawRejected\x18\x0b \x01(\x08\x12\x10\n\x08\x64iameter\x18\x0c \x01(\x01\x12\x15\n\rapriorisample\x18\r \x01(\x01\x12\x13\n\x0b\x61prioriline\x18\x0e \x01(\x01\x12\x13\n\x0bsamplesigma\x18\x0f \x01(\x01\x12\x11\n\tlinesigma\x18\x10 \x01(\x01\x12\x44\n\x03log\x18\x11 \x03(\x0b\x32\x37.Isis.ControlPointFileEntryV0005.Measure.MeasureLogData\x1an\n\x0eMeasureLogData\x12\x16\n\x0e\x64oubleDataType\x18\x01 \x01(\x05\x12\x17\n\x0f\x64oubleDataValue\x18\x02 \x01(\x01\x12\x14\n\x0c\x62oolDataType\x18\x03 \x01(\x05\x12\x15\n\rboolDataValue\x18\x04 \x01(\x08\"U\n\x0bMeasureType\x12\r\n\tCandidate\x10\x00\x12\n\n\x06Manual\x10\x01\x12\x13\n\x0fRegisteredPixel\x10\x02\x12\x16\n\x12RegisteredSubPixel\x10\x03\"X\n\tPointType\x12\x08\n\x04\x46ree\x10\x02\x12\x0f\n\x0b\x43onstrained\x10\x03\x12\t\n\x05\x46ixed\x10\x04\x12\x10\n\x0cobsolete_Tie\x10\x00\x12\x13\n\x0fobsolete_Ground\x10\x01\"\x82\x01\n\rAprioriSource\x12\x08\n\x04None\x10\x00\x12\x08\n\x04User\x10\x01\x12\x15\n\x11\x41verageOfMeasures\x10\x02\x12\r\n\tReference\x10\x03\x12\r\n\tEllipsoid\x10\x04\x12\x07\n\x03\x44\x45M\x10\x05\x12\x0b\n\x07\x42\x61semap\x10\x06\x12\x12\n\x0e\x42undleSolution\x10\x07')
)



_CONTROLPOINTFILEENTRYV0005_MEASURE_MEASURETYPE = _descriptor.EnumDescriptor(
  name='MeasureType',
  full_name='Isis.ControlPointFileEntryV0005.Measure.MeasureType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='Candidate', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Manual', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RegisteredPixel', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RegisteredSubPixel', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1529,
  serialized_end=1614,
)
_sym_db.RegisterEnumDescriptor(_CONTROLPOINTFILEENTRYV0005_MEASURE_MEASURETYPE)

_CONTROLPOINTFILEENTRYV0005_POINTTYPE = _descriptor.EnumDescriptor(
  name='PointType',
  full_name='Isis.ControlPointFileEntryV0005.PointType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='Free', index=0, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Constrained', index=1, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Fixed', index=2, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='obsolete_Tie', index=3, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='obsolete_Ground', index=4, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1616,
  serialized_end=1704,
)
_sym_db.RegisterEnumDescriptor(_CONTROLPOINTFILEENTRYV0005_POINTTYPE)

_CONTROLPOINTFILEENTRYV0005_APRIORISOURCE = _descriptor.EnumDescriptor(
  name='AprioriSource',
  full_name='Isis.ControlPointFileEntryV0005.AprioriSource',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='None', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='User', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='AverageOfMeasures', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Reference', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Ellipsoid', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DEM', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Basemap', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BundleSolution', index=7, number=7,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1707,
  serialized_end=1837,
)
_sym_db.RegisterEnumDescriptor(_CONTROLPOINTFILEENTRYV0005_APRIORISOURCE)


_CONTROLPOINTFILEENTRYV0005_POINTLOGDATA = _descriptor.Descriptor(
  name='PointLogData',
  full_name='Isis.ControlPointFileEntryV0005.PointLogData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='doubleDataType', full_name='Isis.ControlPointFileEntryV0005.PointLogData.doubleDataType', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='doubleDataValue', full_name='Isis.ControlPointFileEntryV0005.PointLogData.doubleDataValue', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='boolDataType', full_name='Isis.ControlPointFileEntryV0005.PointLogData.boolDataType', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='boolDataValue', full_name='Isis.ControlPointFileEntryV0005.PointLogData.boolDataValue', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=860,
  serialized_end=968,
)

_CONTROLPOINTFILEENTRYV0005_MEASURE_MEASURELOGDATA = _descriptor.Descriptor(
  name='MeasureLogData',
  full_name='Isis.ControlPointFileEntryV0005.Measure.MeasureLogData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='doubleDataType', full_name='Isis.ControlPointFileEntryV0005.Measure.MeasureLogData.doubleDataType', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='doubleDataValue', full_name='Isis.ControlPointFileEntryV0005.Measure.MeasureLogData.doubleDataValue', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='boolDataType', full_name='Isis.ControlPointFileEntryV0005.Measure.MeasureLogData.boolDataType', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='boolDataValue', full_name='Isis.ControlPointFileEntryV0005.Measure.MeasureLogData.boolDataValue', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1417,
  serialized_end=1527,
)

_CONTROLPOINTFILEENTRYV0005_MEASURE = _descriptor.Descriptor(
  name='Measure',
  full_name='Isis.ControlPointFileEntryV0005.Measure',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='serialnumber', full_name='Isis.ControlPointFileEntryV0005.Measure.serialnumber', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='Isis.ControlPointFileEntryV0005.Measure.type', index=1,
      number=2, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sample', full_name='Isis.ControlPointFileEntryV0005.Measure.sample', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='line', full_name='Isis.ControlPointFileEntryV0005.Measure.line', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sampleResidual', full_name='Isis.ControlPointFileEntryV0005.Measure.sampleResidual', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lineResidual', full_name='Isis.ControlPointFileEntryV0005.Measure.lineResidual', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='choosername', full_name='Isis.ControlPointFileEntryV0005.Measure.choosername', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='datetime', full_name='Isis.ControlPointFileEntryV0005.Measure.datetime', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='editLock', full_name='Isis.ControlPointFileEntryV0005.Measure.editLock', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ignore', full_name='Isis.ControlPointFileEntryV0005.Measure.ignore', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='jigsawRejected', full_name='Isis.ControlPointFileEntryV0005.Measure.jigsawRejected', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='diameter', full_name='Isis.ControlPointFileEntryV0005.Measure.diameter', index=11,
      number=12, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='apriorisample', full_name='Isis.ControlPointFileEntryV0005.Measure.apriorisample', index=12,
      number=13, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aprioriline', full_name='Isis.ControlPointFileEntryV0005.Measure.aprioriline', index=13,
      number=14, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='samplesigma', full_name='Isis.ControlPointFileEntryV0005.Measure.samplesigma', index=14,
      number=15, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='linesigma', full_name='Isis.ControlPointFileEntryV0005.Measure.linesigma', index=15,
      number=16, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='log', full_name='Isis.ControlPointFileEntryV0005.Measure.log', index=16,
      number=17, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CONTROLPOINTFILEENTRYV0005_MEASURE_MEASURELOGDATA, ],
  enum_types=[
    _CONTROLPOINTFILEENTRYV0005_MEASURE_MEASURETYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=971,
  serialized_end=1614,
)

_CONTROLPOINTFILEENTRYV0005 = _descriptor.Descriptor(
  name='ControlPointFileEntryV0005',
  full_name='Isis.ControlPointFileEntryV0005',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Isis.ControlPointFileEntryV0005.id', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='Isis.ControlPointFileEntryV0005.type', index=1,
      number=2, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chooserName', full_name='Isis.ControlPointFileEntryV0005.chooserName', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='datetime', full_name='Isis.ControlPointFileEntryV0005.datetime', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='editLock', full_name='Isis.ControlPointFileEntryV0005.editLock', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ignore', full_name='Isis.ControlPointFileEntryV0005.ignore', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='jigsawRejected', full_name='Isis.ControlPointFileEntryV0005.jigsawRejected', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='referenceIndex', full_name='Isis.ControlPointFileEntryV0005.referenceIndex', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aprioriSurfPointSource', full_name='Isis.ControlPointFileEntryV0005.aprioriSurfPointSource', index=8,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aprioriSurfPointSourceFile', full_name='Isis.ControlPointFileEntryV0005.aprioriSurfPointSourceFile', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aprioriRadiusSource', full_name='Isis.ControlPointFileEntryV0005.aprioriRadiusSource', index=10,
      number=11, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aprioriRadiusSourceFile', full_name='Isis.ControlPointFileEntryV0005.aprioriRadiusSourceFile', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='latitudeConstrained', full_name='Isis.ControlPointFileEntryV0005.latitudeConstrained', index=12,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='longitudeConstrained', full_name='Isis.ControlPointFileEntryV0005.longitudeConstrained', index=13,
      number=14, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='radiusConstrained', full_name='Isis.ControlPointFileEntryV0005.radiusConstrained', index=14,
      number=15, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aprioriX', full_name='Isis.ControlPointFileEntryV0005.aprioriX', index=15,
      number=16, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aprioriY', full_name='Isis.ControlPointFileEntryV0005.aprioriY', index=16,
      number=17, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aprioriZ', full_name='Isis.ControlPointFileEntryV0005.aprioriZ', index=17,
      number=18, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aprioriCovar', full_name='Isis.ControlPointFileEntryV0005.aprioriCovar', index=18,
      number=19, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='adjustedX', full_name='Isis.ControlPointFileEntryV0005.adjustedX', index=19,
      number=20, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='adjustedY', full_name='Isis.ControlPointFileEntryV0005.adjustedY', index=20,
      number=21, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='adjustedZ', full_name='Isis.ControlPointFileEntryV0005.adjustedZ', index=21,
      number=22, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='adjustedCovar', full_name='Isis.ControlPointFileEntryV0005.adjustedCovar', index=22,
      number=23, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='log', full_name='Isis.ControlPointFileEntryV0005.log', index=23,
      number=24, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='measures', full_name='Isis.ControlPointFileEntryV0005.measures', index=24,
      number=25, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CONTROLPOINTFILEENTRYV0005_POINTLOGDATA, _CONTROLPOINTFILEENTRYV0005_MEASURE, ],
  enum_types=[
    _CONTROLPOINTFILEENTRYV0005_POINTTYPE,
    _CONTROLPOINTFILEENTRYV0005_APRIORISOURCE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=43,
  serialized_end=1837,
)

_CONTROLPOINTFILEENTRYV0005_POINTLOGDATA.containing_type = _CONTROLPOINTFILEENTRYV0005
_CONTROLPOINTFILEENTRYV0005_MEASURE_MEASURELOGDATA.containing_type = _CONTROLPOINTFILEENTRYV0005_MEASURE
_CONTROLPOINTFILEENTRYV0005_MEASURE.fields_by_name['type'].enum_type = _CONTROLPOINTFILEENTRYV0005_MEASURE_MEASURETYPE
_CONTROLPOINTFILEENTRYV0005_MEASURE.fields_by_name['log'].message_type = _CONTROLPOINTFILEENTRYV0005_MEASURE_MEASURELOGDATA
_CONTROLPOINTFILEENTRYV0005_MEASURE.containing_type = _CONTROLPOINTFILEENTRYV0005
_CONTROLPOINTFILEENTRYV0005_MEASURE_MEASURETYPE.containing_type = _CONTROLPOINTFILEENTRYV0005_MEASURE
_CONTROLPOINTFILEENTRYV0005.fields_by_name['type'].enum_type = _CONTROLPOINTFILEENTRYV0005_POINTTYPE
_CONTROLPOINTFILEENTRYV0005.fields_by_name['aprioriSurfPointSource'].enum_type = _CONTROLPOINTFILEENTRYV0005_APRIORISOURCE
_CONTROLPOINTFILEENTRYV0005.fields_by_name['aprioriRadiusSource'].enum_type = _CONTROLPOINTFILEENTRYV0005_APRIORISOURCE
_CONTROLPOINTFILEENTRYV0005.fields_by_name['log'].message_type = _CONTROLPOINTFILEENTRYV0005_POINTLOGDATA
_CONTROLPOINTFILEENTRYV0005.fields_by_name['measures'].message_type = _CONTROLPOINTFILEENTRYV0005_MEASURE
_CONTROLPOINTFILEENTRYV0005_POINTTYPE.containing_type = _CONTROLPOINTFILEENTRYV0005
_CONTROLPOINTFILEENTRYV0005_APRIORISOURCE.containing_type = _CONTROLPOINTFILEENTRYV0005
DESCRIPTOR.message_types_by_name['ControlPointFileEntryV0005'] = _CONTROLPOINTFILEENTRYV0005
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ControlPointFileEntryV0005 = _reflection.GeneratedProtocolMessageType('ControlPointFileEntryV0005', (_message.Message,), dict(

  PointLogData = _reflection.GeneratedProtocolMessageType('PointLogData', (_message.Message,), dict(
    DESCRIPTOR = _CONTROLPOINTFILEENTRYV0005_POINTLOGDATA,
    __module__ = 'ControlPointFileEntryV0005_pb2'
    # @@protoc_insertion_point(class_scope:Isis.ControlPointFileEntryV0005.PointLogData)
    ))
  ,

  Measure = _reflection.GeneratedProtocolMessageType('Measure', (_message.Message,), dict(

    MeasureLogData = _reflection.GeneratedProtocolMessageType('MeasureLogData', (_message.Message,), dict(
      DESCRIPTOR = _CONTROLPOINTFILEENTRYV0005_MEASURE_MEASURELOGDATA,
      __module__ = 'ControlPointFileEntryV0005_pb2'
      # @@protoc_insertion_point(class_scope:Isis.ControlPointFileEntryV0005.Measure.MeasureLogData)
      ))
    ,
    DESCRIPTOR = _CONTROLPOINTFILEENTRYV0005_MEASURE,
    __module__ = 'ControlPointFileEntryV0005_pb2'
    # @@protoc_insertion_point(class_scope:Isis.ControlPointFileEntryV0005.Measure)
    ))
  ,
  DESCRIPTOR = _CONTROLPOINTFILEENTRYV0005,
  __module__ = 'ControlPointFileEntryV0005_pb2'
  # @@protoc_insertion_point(class_scope:Isis.ControlPointFileEntryV0005)
  ))
_sym_db.RegisterMessage(ControlPointFileEntryV0005)
_sym_db.RegisterMessage(ControlPointFileEntryV0005.PointLogData)
_sym_db.RegisterMessage(ControlPointFileEntryV0005.Measure)
_sym_db.RegisterMessage(ControlPointFileEntryV0005.Measure.MeasureLogData)


_CONTROLPOINTFILEENTRYV0005.fields_by_name['aprioriCovar']._options = None
_CONTROLPOINTFILEENTRYV0005.fields_by_name['adjustedCovar']._options = None
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
import sys
import math
import numpy as np


def zero360(angles, rad=False):
    """
    Convert angle to base 0-360

    Parameters
    ----------
    angles : float
             a scalar angle to convert

    rad : boolean
          flag whether angles are in radians

    Returns
    -------
    bb : float
         converted angle
    """

    if rad:
        half = math.pi
    else:
        half = 180.0

    circle = 2.0 * half

    ii = np.floor(angles / circle)
    bb = angles - circle * ii
    #Negative values to positive
    bb[bb < 0] += circle
    return bb


def julian2ls(date, marsyear=None, reverse=False):
    """

    Original IDL from Hugh Keiffer

    Parameters
    -----------
    date : numeric
           Scalar or NumPy ndarray or dates

    marsyear : float
              Mars year for use in reverse

    reverse : bool
              Reverse conversion from L_{s} to julian

    Returns
    -------
    out : numeric
          float or array of float LsubS or Julian dates

    myn : float
          If LsubS to Mars year, return the Mars year

    References
    -----------
    [1] M. Allison and M. McEwen.' A post-Pathfinder evaluation of areocentric
    solar coordinates with improved timing recipes for Mars seasonal/diurnal
    climate studies'. Plan. Space Sci., 2000, v=48, pages = {215-235},

    [2] http://www.giss.nasa.gov/tools/mars24/help/algorithm.html

    """

    if not isinstance(date, np.ndarray):
        date = np.asarray([date], dtype=np.float64)

    rec= np.array([-0.0043336633,0.0043630568,0.0039601861,0.029025116,-0.00042300026])
    dj2000 = 2451545.0  # JD of epoch J2000
    dj4m = 51544.50  # days offset from dj4 to djm
    smja = 1.523679  # semi-major axis in AU, Table 2. last digit from Ref2
    meanmotion = 0.52402075  # mean motion, degrees/day. Constant in Table 2

    a5 = np.array([71,57,39,37,21,20,18]) * 0.0001
    tau = np.array([2.2353,2.7543,1.1177,15.7866,2.1354,2.4694,32.8493])
    phi = np.array([49.409,168.173,191.837,21.736,15.704,95.528,49.095])


    if reverse:
        #Convert from lsubs to julian
        lsubs = date
        delta_lsubs = lsubs - 250.99864
        rdelta_lsubs= np.radians(delta_lsubs)
        dj4 = 51507.5 + 1.90826 * delta_lsubs -\
            20.42 * np.sin(rdelta_lsubs) +\
            0.72 * np.sin( 2. * rdelta_lsubs)  # Eq. 14

        brak = 686.9726 + 0.0043 * np.cos(rdelta_lsubs) -\
            0.0003 * np.cos(2. * rdelta_lsubs) # in brackets in Eq 14
        if marsyear is None:
            return False
        ny=marsyear + 42  # orbits of mars since 1874.0
        dj4 += brak * (ny - 66)  # last part of Eq. 14
        djtt= dj4 - dj4m  # days from J2000 TT
        tcen = djtt / 36525.0  # julian centuries from J2000
        tcor= 64.184 + tcen * (59.0 + tcen * (-51.2 + tcen * (-67.1 - tcen * 16.4)))  # TT-UTC
        djm=djtt - tcor / 86400.0  # convert correction to days and apply
        out=djm  #  out is UTC
    else:
        #Convert from julian to lsubs
        if date[0] < 1.e6:  #Is this the intended functionality - what if date[0] < and date [1] >
            djm = date
        else:
            djm = date - dj2000
        tcen = djm / 36525.0  # julian centuries from J2000
        tcor = 64.184 + tcen * (59.0 + tcen * (-51.2 + tcen * (-67.1 - tcen * 16.4)))  # TT-UTC
        djtt = djm + tcor / 86400.0  # convert correction to days and apply. get TT
        dj4 = djtt + dj4m  #  A+M's MJD
    nelements = date.size
    pbs = np.zeros(nelements)
    for i in range(nelements):
        q = 0.985626 * djtt[i]
        rq = np.radians(q / tau + phi)
        pbsi = a5 * np.cos(rq)
        pbs[i] = pbsi.sum()
    meananomoly = 19.3870 + meanmotion * djtt  # Mean anomoly M, Table 2 and Eq. 16
    rmeananomoly = np.radians(meananomoly)  # M in radians
    #Eqn of center, in brackets in eq 20 and all but the first term in eq 19. exclude pbs
    eoc = (10.691 + 3.e-7 * djtt) * np.sin(rmeananomoly)\
        + 0.623 * np.sin(2. * rmeananomoly) + 0.050 * np.sin(3. * rmeananomoly)\
        + 0.005 * np.sin(4. * rmeananomoly) + 0.0005 * np.sin(5. * rmeananomoly)

    if reverse:
        try:
            j = reverse.size
        except:
            j=0
            rev = [0]
        if j >= 5:
            #User has supplied coefficients
            rec = rev
        if j < 5 and rev[0] < 5:
            fd = 0.0
        else:
            fd=rec[0] + rec[1]* np.sin(rdelta_lsubs)\
                + rec[2] * np.sin(2. * rdelta_lsubs)\
                + rec[3] * np.sin(3. * rdelta_lsubs)\
                + rec[4] * np.sin(4. * rdelta_lsubs)
        #TODO: This looks like reversing is going to cause an fd undefined error if j >= 5
        out = out - ( pbs / meananomoly + fd)
    else:
        afms = 270.3863 + 0.52403840 * djtt # Eq. 17
        lsubs = afms + eoc + pbs              # Eq. 19 LS in degrees
        lsubs = zero360(lsubs)
        out= lsubs
        marstyear = 686.9728  # mean mars tropical  year in terrestrial days
        #marsysol=668.5991  #mars siderial year in sols
        ny = np.floor((dj4 - 5668.690)/ marstyear ) # full Mars years from 1874
        myn = ny - 42  # climate MY

    if reverse:
        return out
    else:
        return out, myn

if __name__ == '__main__':
    print(julian2ls(178, marsyear=40, reverse=True))




import unittest

from plio.io import io_json
from plio.io import io_yaml

try:
    import yaml
    missing = False
except ImportError:
    missing = True

from plio.examples import get_path


class TestYAML(unittest.TestCase):

    @unittest.skipIf(missing == True, 'YAML library not installed')
    def test_read(self):
        d = io_yaml.read_yaml(get_path('logging.yaml'))
        self.assertIn('handlers', d.keys())


class TestJSON(unittest.TestCase):

    def test_read(self):
        d = io_json.read_json(get_path('logging.json'))
        self.assertIn('handlers', d.keys())



import pvl
from plio.utils.utils import find_in_dict

def extract_keywords(header, *args):
    """
    For a given header, find all of the keys and return an unnested dict.
    """
    try:
        header = pvl.load(header)
    except:
        header = pvl.loads(header)

    res = {}
    # Iterate through all of the requested keys
    for a in args:
        try:
            res[a] = find_in_dict(a)
        except:
            res[a] = None
    return res



# Conditional imports for GDAL
import importlib
import warnings
import sys

try:
    gdal = importlib.util.find_spec('gdal')
    ogr = importlib.util.find_spec('osgeo.ogr')
    osr = importlib.util.find_spec('osr')

    gdal = gdal.loader.load_module()
    ogr = ogr.loader.load_module()
    osr = osr.loader.load_module()
    gdal.UseExceptions() 
except:
    try:
        gdal = importlib.util.find_spec('osgeo.gdal')
        gdal = gdal.loader.load_module()
    except:
        gdal = None
    ogr = None
    osr = None

def conditional_gdal(func):
    def has_gdal(*args, **kwargs):
        if gdal:
            return func(*args, **kwargs)
        else:
            warnings.warn('Trying to call a GDAL method, but GDAL is not installed.')
        return None
    return has_gdal

from . import io_autocnetgraph
from . import io_controlnetwork
from . import io_db
from . import io_gdal
from . import io_hdf
from . import io_json
from . import io_krc
from . import io_pvl
from . import io_spectral_profiler
from . import io_tes
from . import io_yaml
from . import isis_serial_number



import numpy as np
import unittest
import datetime

from .. import marstime

class TestMarsTime(unittest.TestCase):
    def equality_test(self):
        earth = marstime.getUTCfromLS(24, 130)
        mars = marstime.getMTfromTime(earth)
        assert(earth.date() == marstime.getUTCfromLS(mars.year,mars.ls).date())

    def test_J2000(self):
        iTime = [2001,11,13,2,45,2]
        testJD = marstime.getJ2000(iTime)
        callibration = 58891502.000000 #test should be this value.
        diff = testJD - callibration
        print(testJD)
        print(diff)
        self.assertTrue(diff == -58890820.38465065)


    def test_UTC(self):
        jd = 2452226.614606
        date = marstime.getUTC(jd)
        callibration = datetime.datetime(2001,11,13,2,45,2)
        diff = callibration - date
        self.assertTrue(str(diff) == '0:00:00.041599')


    def test_LS(self):
        iTime = [2000,1,6,0,0,0]
        lsdata = marstime.getMTfromTime(iTime)
        ls = lsdata.ls
        year = lsdata.year
        callibration = 277.18677
        diff = ls - callibration
        print(diff)
        self.assertAlmostEqual(diff, 0.0)

    def test_SZA(self):
        '''test getSZAfromTime'''
        itime = [2000,1,6,0,0,0]
        lon = 0.0
        lat = 0.0
        expected = 154.261908004
        sza = marstime.getSZAfromTime(itime,lon,lat)
        self.assertAlmostEqual(sza, expected)

    def test_SZAGetTime(self):
        out = marstime.SZAGetTime(0, [1,1,1], 0,0)
        self.assertAlmostEqual(out[1], 9.454213735250395)


    def test_LTfromTime(self):
        '''test getLTfromTime function'''
        iTime = [2000,1,6,0,0,0]
        lon = 0.0
        LTST = marstime.getLTfromTime(iTime,lon)
        expected = 23.648468934
        self.assertAlmostEqual(LTST, expected)
